# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
from enum import Enum
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ...constants import SMALL, RE, J2, J4, MU, TWOPI
from .frame_conversions import rv2coe, coe2rv, rv2pqw
from .newton import newtonm
from .utils import is_equatorial, findc2c3


logger = logging.getLogger(__name__)


class FGMethod(Enum):
    PQW = "pqw"
    SERIES = "series"
    C2C3 = "c2c3"


def kepler(
    ro: ArrayLike, vo: ArrayLike, dtsec: float, n_iters: int = 50
) -> Tuple[np.ndarray, np.ndarray]:
    """Solves Kepler's problem for orbit determination and returns a future geocentric
    equatorial (ECI) position and velocity vector using universal variables.

    References:
        Vallado: 2022, p. 94-96, Algorithm 8

    Args:
        ro (array_like): Initial ECI position vector in km
        vo (array_like): Initial ECI velocity vector in km/s
        dtsec (float): Time interval to propagate in seconds
        n_iters (int, optional): Number of iterations for Newton-Raphson
                                 method

    Returns:
        tuple: (r, v)
            r (np.ndarray): Propagated ECI position vector in km
            v (np.ndarray): Propagated ECI velocity vector in km/s
    """
    # Convert to numpy arrays
    ro, vo = np.array(ro), np.array(vo)

    # Initialize values
    ktr = 0
    xnew, znew, c2new, c3new = 0, 0, 0, 0
    dtnew = -10
    smu = np.sqrt(MU)
    magro = np.linalg.norm(ro)
    magvo = np.linalg.norm(vo)
    rdotv = np.dot(ro, vo)

    # Find specific mechanical energy, alpha, and semi-major axis
    sme = (magvo**2 / 2) - (MU / magro)
    alpha = -2 * sme / MU
    a = -MU / (2 * sme) if np.abs(sme) > SMALL else np.inf
    alpha = 0 if np.abs(alpha) < SMALL else alpha

    # Setup initial guess for x
    if alpha >= SMALL:
        # Circular and elliptical orbits
        period = TWOPI * np.sqrt(np.abs(a) ** 3 / MU)
        if np.abs(dtsec) > np.abs(period):
            dtsec = dtsec % period
        xold = smu * dtsec * alpha
    elif np.abs(alpha) < SMALL:
        # Parabolic orbit
        h = np.cross(ro, vo)
        magh = np.linalg.norm(h)
        p = magh**2 / MU
        s = 0.5 * (np.pi / 2 - np.arctan(3 * np.sqrt(MU / (p**3)) * dtsec))
        w = np.arctan(np.tan(s) ** (1 / 3))
        xold = np.sqrt(p) * (2 / np.tan(2 * w))
        alpha = 0
    else:
        # Hyperbolic orbit
        temp = (
            -2
            * MU
            * dtsec
            / (a * (rdotv + np.sign(dtsec) * np.sqrt(-MU * a) * (1 - magro * alpha)))
        )
        xold = np.sign(dtsec) * np.sqrt(-a) * np.log(temp)

    # Newton-Raphson iteration to find x
    tmp = 1 / smu
    while (np.abs(dtnew * tmp - dtsec) >= SMALL) and (ktr < n_iters):
        xoldsqrd = xold * xold
        znew = xoldsqrd * alpha

        # Find c2 and c3 functions
        c2new, c3new = findc2c3(znew)

        # Use a newton iteration for new values
        rval = (
            xoldsqrd * c2new
            + rdotv * tmp * xold * (1 - znew * c3new)
            + magro * (1 - znew * c2new)
        )
        dtnew = (
            xoldsqrd * xold * c3new
            + rdotv * tmp * xoldsqrd * c2new
            + magro * xold * (1 - znew * c3new)
        )

        # Calculate new value for x
        temp1 = (dtsec * smu - dtnew) / rval
        xnew = xold + temp1

        # Check if the univ param goes negative; if so, use bissection
        if (xnew < 0) and (dtsec > 0):
            xnew = xold * 0.5

        ktr += 1
        xold = xnew

    # Check for convergence
    if ktr >= n_iters:
        logger.error(
            f"Kepler not converged in {n_iters} iterations for dtsec = {dtsec}"
        )
        return np.zeros(3), np.zeros(3)

    # Find f and g values
    xnewsqrd = xnew * xnew
    f = 1 - (xnewsqrd * c2new / magro)
    g = dtsec - xnewsqrd * xnew * c3new / smu

    # Find position and velocity vectors at new time
    r = f * ro + g * vo
    magr = np.linalg.norm(r)
    gdot = 1 - (xnewsqrd * c2new / magr)
    fdot = (smu * xnew / (magro * magr)) * (znew * c3new - 1)
    v = fdot * ro + gdot * vo

    # Check if f and g values are consistent
    temp = f * gdot - fdot * g
    if np.abs(temp - 1) > 1e-5:
        logger.warning("f and g values are inconsistent")

    return r, v


def _process_coe(ro, vo):
    # Convert position and velocity to orbital elements
    output = rv2coe(ro, vo)
    processed_output = tuple(
        0 if isinstance(x, float) and np.isnan(x) else x for x in output
    )
    p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper, _ = processed_output

    # Check for negative semi-major axis and set mean motion
    if a < 0:
        logger.error("Negative semi-major axis encountered")
        n = None
    else:
        n = np.sqrt(MU / (a**3))

    return p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper, n


def _calc_rv_from_coe(
    incl,
    truelon,
    raandot,
    argpdot,
    mdot,
    ndot,
    nddot,
    ecc,
    p,
    m,
    dtsec,
    raan,
    argp,
    arglat,
    lonper,
    nu,
):
    # Update orbital elements
    if ecc < SMALL:
        # Circular orbit
        if is_equatorial(incl):
            # Circular equatorial
            truelon += (raandot + argpdot + mdot) * dtsec
            truelon = np.mod(truelon, TWOPI)
        else:
            # Circular inclined
            raan += raandot * dtsec
            raan = np.mod(raan, TWOPI)
            arglat += (argpdot + mdot) * dtsec
            arglat = np.mod(arglat, TWOPI)
    else:
        # Elliptical orbit
        if is_equatorial(incl):
            # Elliptical equatorial
            lonper += (raandot + argpdot) * dtsec
            lonper = np.mod(lonper, TWOPI)
        else:
            # Elliptical inclined
            raan += raandot * dtsec
            raan = np.mod(raan, TWOPI)
            argp += argpdot * dtsec
            argp = np.mod(argp, TWOPI)

        m += mdot * dtsec + ndot * dtsec**2 + nddot * dtsec**3
        m = np.mod(m, TWOPI)
        e0, nu = newtonm(ecc, m)

    # Convert updated orbital elements back to position and velocity vectors
    r, v = coe2rv(p, ecc, incl, raan, argp, nu, arglat, truelon, lonper)

    return r, v


def pkepler(
    ro: ArrayLike, vo: ArrayLike, dtsec: float, ndot: float, nddot: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Propagates a satellite's position and velocity vectors over a given time period
    accounting for J2 perturbations.

    References:
        Vallado: 2022, p. 715-717, Algorithm 66

    Args:
        ro (array_like): Initial ECI position vector in km
        vo (array_like): Initial ECI velocity vector in km/s
        dtsec (float): Time interval to propagate in seconds
        ndot (float): First time derivative of mean motion in rad/s²
        nddot (float): Second time derivative of mean motion in rad/s³

    Returns:
        tuple: (r, v)
            r (np.ndarray): Propagated ECI position vector in km
            v (np.ndarray): Propagated ECI velocity vector in km/s

    TODO:
        - Move to perturbations?
    """
    # Convert position and velocity to orbital elements
    p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper, n = _process_coe(
        ro, vo
    )

    # Check for negative semi-major axis
    if n is None:
        return np.zeros(3), np.zeros(3)

    # J2 perturbation effects
    j2op2 = (n * 1.5 * RE**2 * J2) / (p**2)
    raandot = -j2op2 * np.cos(incl)
    argpdot = j2op2 * (2 - 2.5 * np.sin(incl) ** 2)
    mdot = n

    # Update semi-major axis and eccentricity
    a -= 2 * ndot * dtsec * a / (3 * n)
    ecc -= 2 * (1 - ecc) * ndot * dtsec / (3 * n)
    p = a * (1 - ecc**2)

    # Calculate position and velocity vectors
    return _calc_rv_from_coe(
        incl,
        truelon,
        raandot,
        argpdot,
        mdot,
        ndot,
        nddot,
        ecc,
        p,
        m,
        dtsec,
        raan,
        argp,
        arglat,
        lonper,
        nu,
    )


def pkeplerj4(
    ro: ArrayLike, vo: ArrayLike, dtsec: float, ndot: float, nddot: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Propagates a satellite's position and velocity vector over a given time period
    accounting for J2, J2^2, and J4 perturbation effects.

    References:
        Vallado: 2022, p. 715-717

    Args:
        ro (array_like): Initial ECI position vector in km
        vo (array_like): Initial ECI velocity vector in km/s
        dtsec (float): Time interval to propagate in seconds
        ndot (float): First time derivative of mean motion in rad/s²
        nddot (float): Second time derivative of mean motion in rad/s³

    Returns:
        tuple: (r, v)
            r (np.ndarray): Propagated ECI position vector in km
            v (np.ndarray): Propagated ECI velocity vector in km/s

    TODO:
        - Move to perturbations?
    """
    # Convert position and velocity to orbital elements
    p, a, ecc, incl, raan, argp, nu, m, arglat, truelon, lonper, n = _process_coe(
        ro, vo
    )

    # Check for negative semi-major axis
    if n is None:
        return np.zeros(3), np.zeros(3)

    # Intermediate calculations
    cosi, sini = np.cos(incl), np.sin(incl)
    beta2 = 1 - ecc**2
    sqrtbeta = np.sqrt(beta2)

    # J2 perturbations calculations
    nbar = n * (
        1.5 * J2 * (RE / p) ** 2 * sqrtbeta * (1 - 1.5 * sini**2)
        + 3
        / 128
        * J2**2
        * (RE / p) ** 4
        * sqrtbeta
        * (
            16 * sqrtbeta
            + 25 * beta2
            - 15
            + (30 - 96 * sqrtbeta - 90 * beta2) * cosi**2
            + (105 + 144 * sqrtbeta + 25 * beta2) * cosi**4
        )
        - 45
        / 128
        * J4
        * ecc**2
        * (RE / p) ** 4
        * sqrtbeta
        * (3 - 30 * cosi**2 + 35 * cosi**4)
    )

    mdot = n + nbar

    # RAAN and argument of perigee perturbation rates
    raandot = (
        -1.5 * J2 * (RE / p) ** 2 * mdot * cosi
        - (9 / 96)
        * J2**2
        * (RE / p) ** 4
        * mdot
        * cosi
        * (
            36
            + 4 * ecc**2
            - 48 * sqrtbeta
            - (40 - 5 * ecc**2 - 72 * sqrtbeta) * sini * sini
        )
        - (35 / 112)
        * J4
        * (RE / p) ** 4
        * n
        * cosi
        * (1 + 1.5 * ecc**2)
        * (12 - 21 * sini**2)
    )

    # Argument of perigee rate calculations
    argpdot = (
        0.75 * J2 * (RE / p) ** 2 * mdot * (4 - 5 * sini**2)
        + (9 / 192)
        * J2**2
        * (RE / p) ** 4
        * mdot
        * (2 - 2.5 * sini**2)
        * (
            96
            + 24 * ecc**2
            - 96 * sqrtbeta
            - (86 - ecc**2 - 144 * sqrtbeta) * sini * sini
        )
        - (45 / 36) * J2**2 * (RE / p) ** 4 * ecc**2 * n * cosi**4
        - (35 / 896)
        * J4
        * (RE / p) ** 4
        * n
        * (
            192
            - 744 * sini**2
            + 588 * sini**4
            + ecc**2 * (216 - 756 * sini**2 + 567 * sini**4)
        )
    )

    # Calculate position and velocity vectors
    return _calc_rv_from_coe(
        incl,
        truelon,
        raandot,
        argpdot,
        mdot,
        ndot,
        nddot,
        ecc,
        p,
        m,
        dtsec,
        raan,
        argp,
        arglat,
        lonper,
        nu,
    )


def findfandg(
    r1: ArrayLike,
    v1: ArrayLike,
    r2: ArrayLike,
    v2: ArrayLike,
    dtsec: float,
    x: float,
    z: float,
    c2: float,
    c3: float,
    method: FGMethod,
) -> Tuple[float, float, float, float]:
    """Calculates the f and g functions for use in various applications.

    References:
        Vallado: 2022, p. 83-88

    Args:
        r1 (array_like): First position vector in km
        v1 (array_like): First velocity vector in km/s
        r2 (array_like): Second position vector in km
        v2 (array_like): Second velocity vector in km/s
        dtsec (float): Step size in seconds
        x (float): Universal variable x
        z (float): Universal variable z
        c2 (float): c2 function value
        c3 (float): c3 function value
        method (FGMethod): Method to use for calculating f and g values

    Returns:
        tuple: (f, g, fdot, gdot)
            f (float): f function value
            g (float): g function value
            fdot (float): fdot function value
            gdot (float): gdot function value

    Notes:
        - The step size `dtsec` should be small (on the order of 60-120 seconds)
    """
    f = g = fdot = gdot = 0
    magr1, magv1 = np.linalg.norm(r1), np.linalg.norm(v1)

    if method == FGMethod.PQW:
        hbar = np.cross(r1, v1)
        h = np.linalg.norm(hbar)
        rpqw1, vpqw1 = rv2pqw(r1, v1)
        rpqw2, vpqw2 = rv2pqw(r2, v2)

        f = (rpqw2[0] * vpqw1[1] - vpqw2[0] * rpqw1[1]) / h
        g = (rpqw1[0] * rpqw2[1] - rpqw2[0] * rpqw1[1]) / h
        gdot = (rpqw1[0] * vpqw2[1] - vpqw2[0] * rpqw1[1]) / h
        fdot = (vpqw2[0] * vpqw1[1] - vpqw2[1] * vpqw1[0]) / h

    elif method == FGMethod.SERIES:
        u = MU / (magr1**3)
        p = np.dot(r1, v1) / (magr1**2)
        q = (magv1**2 - u * magr1**2) / (magr1**2)

        p2, p4, p6 = p**2, p**4, p**6
        u2, u3 = u**2, u**3
        q2, q3 = q**2, q**3
        dt2, dt3, dt4, dt5, dt6, dt7, dt8 = [dtsec**i for i in range(2, 9)]

        # fmt: off
        f = (
            1 - 0.5 * u * dt2 + 0.5 * u * p * dt3
            + u / 24 * (-15 * p2 + 3 * q + u) * dt4
            + p * u / 8 * (7 * p2 - 3 * q - u) * dt5
            + u / 720
            * (-945 * p4 + 630 * p2 * q + 210 * u * p2 - 45 * q2 - 24 * u * q - u2)
            * dt6
            + p * u / 80
            * (165 * p4 - 150 * p2 * q - 50 * u * p2 + 25 * q2 + 14 * u * q + u2) * dt7
            + u / 40320
            * (
                -135135 * p6
                + 155925 * p4 * q
                + 51975 * u * p4
                - 42525 * p2 * q2
                - 24570 * u * p2 * q
                - 2205 * u2 * p2
                + 1575 * q3
                + 1107 * u * q2
                + 117 * u2 * q
                + u3
            )
            * dt8
        )

        g = (
            dtsec - 1 / 6 * u * dt3 + 0.25 * u * p * dt4
            + u / 120 * (-45 * p2 + 9 * q + u) * dt5
            + p * u / 24 * (14 * p2 - 6 * q - u) * dt6
            + u / 5040
            * (-4725 * p4 + 3150 * p2 * q + 630 * u * p2 - 225 * q2 - 54 * u * q - u2)
            * dt7
            + p * u / 320
            * (495 * p4 - 450 * p2 * q - 100 * u * p2 + 75 * q2 + 24.0 * u * q + u2)
            * dt8
        )

        fdot = (
            -u * dtsec + 1.5 * u * p * dt2
            + u / 6 * (-15 * p2 + 3 * q + u) * dt3
            + 5 * p * u / 8 * (7 * p2 - 3 * q - u) * dt4
            + u / 120
            * (-945 * p4 + 630 * p2 * q + 210 * u * p2 - 45 * q2 - 24 * u * q - u2)
            * dt5
            + 7 * p * u / 80
            * (165 * p4 - 150 * p2 * q - 50 * u * p2 + 25 * q2 + 14 * u * q + u2) * dt6
            + u / 5040
            * (-135135 * p6 + 155925 * p4 * q + 51975 * u * p4 - 42525 * p2 * q2
               - 24570 * u * p2 * q - 2205 * u2 * p2 + 1575 * q3 + 1107 * u * q2
               + 117 * u2 * q + u3) * dt7
        )

        gdot = (
            1 - 0.5 * u * dt2 + u * p * dt3
            + u / 24 * (-45 * p2 + 9 * q + u) * dt4
            + p * u / 4 * (14 * p2 - 6 * q - u) * dt5
            + u / 720
            * (-4725 * p4 + 3150 * p2 * q + 630 * u * p2 - 225 * q2 - 54 * u * q - u2)
            * dt6
            + p * u / 40
            * (495 * p4 - 450 * p2 * q - 100 * u * p2 + 75 * q2 + 24 * u * q + u2) * dt7
        )
        # fmt: on

    elif method == FGMethod.C2C3:
        xsqrd = x**2
        magr2 = np.linalg.norm(r2)
        f = 1 - (xsqrd * c2 / magr1)
        g = dtsec - xsqrd * x * c3 / np.sqrt(MU)
        gdot = 1 - (xsqrd * c2 / magr2)
        fdot = (np.sqrt(MU) * x / (magr1 * magr2)) * (z * c3 - 1)

    return f, g, fdot, gdot



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
import os
from dataclasses import dataclass
from pathlib import Path

import numpy as np

from ...constants import ARCSEC2RAD, JD_TO_MJD_OFFSET
from ...mathtime.julian_date import jday


logger = logging.getLogger(__name__)


# Default data directory
ROOT_DIR = Path(__file__).resolve().parents[6]
DATA_DIR = ROOT_DIR / "datalib"

# Conversion factors
CONVRTU = 1e-6 * ARCSEC2RAD  # microarcseconds to radians
CONVRTM = 1e-3 * ARCSEC2RAD  # milliarcseconds to radians


@dataclass
class IAU80Array:
    # fmt: off
    """Data class for IAU 1980 nutation data."""
    iar80: np.ndarray = None
    rar80: np.ndarray = None


@dataclass
class IAU06pnOldArray:
    # fmt: off
    """Data class for IAU 2006 precession-nutation data (older version)."""
    apn: np.ndarray = None
    apni: np.ndarray = None
    appl: np.ndarray = None
    appli: np.ndarray = None


@dataclass
class IAU06Array:
    # fmt: off
    """Data class for IAU 2006 precession-nutation data."""
    ax0: np.ndarray = None
    ax0i: np.ndarray = None
    ay0: np.ndarray = None
    ay0i: np.ndarray = None
    as0: np.ndarray = None
    as0i: np.ndarray = None
    agst: np.ndarray = None
    agsti: np.ndarray = None
    apn0: np.ndarray = None
    apn0i: np.ndarray = None
    apl0: np.ndarray = None
    apl0i: np.ndarray = None
    aapn0: np.ndarray = None
    aapn0i: np.ndarray = None


@dataclass
class IAU06xysArray:
    # fmt: off
    """Data class for IAU 2006 XYS data."""
    jd: np.ndarray = None
    jdf: np.ndarray = None
    x: np.ndarray = None
    y: np.ndarray = None
    s: np.ndarray = None
    mjd_tt: np.ndarray = None


@dataclass
class EOPArray:
    # fmt: off
    """Data class for Earth Orientation Parameters."""
    mjd: np.ndarray = None
    xp: np.ndarray = None
    yp: np.ndarray = None
    dut1: np.ndarray = None
    lod: np.ndarray = None
    ddpsi: np.ndarray = None
    ddeps: np.ndarray = None
    dx: np.ndarray = None
    dy: np.ndarray = None
    dat: np.ndarray = None


@dataclass
class SPWArray:
    # fmt: off`
    """Data class for Space Weather data."""
    kparray: np.ndarray = None
    sumkp: np.ndarray = None
    aparray: np.ndarray = None
    avgap: np.ndarray = None
    adjf107: np.ndarray = None
    adjctrf81: np.ndarray = None
    adjlstf81: np.ndarray = None
    obsf107: np.ndarray = None
    obsctrf81: np.ndarray = None
    obslstf81: np.ndarray = None
    mjd: np.ndarray = None


def iau80in(data_dir: str = DATA_DIR) -> IAU80Array:
    """Initializes the nutation matrices needed for reduction calculations.

    References:
        Vallado, 2022, Section 3.7.1

    Args:
        data_dir (str, optional): Directory containing the nutation data file
                                  "nut80.dat" (default: DATA_DIR)

    Returns:
        IAU80Array: Data object containing the nutation matrices
            iar80 (np.ndarray): Integers for FK5 1980
            rar80 (np.ndarray): Reals for FK5 1980 in radians
    """
    # Load the nutation data and initialize data object
    nut80 = np.loadtxt(os.path.join(data_dir, "nut80.dat"))
    iau80arr = IAU80Array()

    # Split into integer and real parts
    iau80arr.iar80 = nut80[:, :5].astype(int)
    iau80arr.rar80 = nut80[:, 5:9]

    # Convert from 0.0001 arcseconds to radians
    convrt = 1e-4 * ARCSEC2RAD
    iau80arr.rar80 *= convrt

    return iau80arr


def iau06in_pnold(data_dir: str = DATA_DIR) -> IAU06pnOldArray:
    """Initializes the nutation matrices needed for IAU 2006 reduction calculations.

    This is an older version of the function that loads the data from the IAU 2006,
    and we're only returning the coefficients for nutation.

    References:
        Vallado, 2022, Section 3.7.1

    Returns:
        IAU06pnOldArray: Data object containing the nutation coefficients
            apn (np.ndarray): Nutation coefficients for IAU 2006
            apni (np.ndarray): Integers for nutation coefficients
            appl (np.ndarray): Planetary nutation coefficients for IAU 2006
            appli (np.ndarray): Integers for planetary nutation coefficients

    Notes:
        Data files are from the IAU 2006 precession-nutation model:
            - iau03n.dat (file for nutation coefficients)
            - iau03pl.dat (file for planetary nutation coefficients)

    TODO:
        - Deprecate when updates to iau06pna/b are complete.
    """

    def load_data(
        filename, columns_real, columns_int, conv_factor, convert_exclude_last=False
    ):
        """Helper function to load and process data."""
        filepath = os.path.join(data_dir, filename)
        data = np.loadtxt(filepath)
        reals = data[:, columns_real]
        if convert_exclude_last:
            reals[:, :-1] *= conv_factor  # convert all except the last column
        else:
            reals *= conv_factor  # convert all
        integers = data[:, columns_int].astype(int)
        return reals, integers

    # Initialize data object
    iau06arr = IAU06pnOldArray()

    # Load all data files
    iau06arr.apn, iau06arr.apni = load_data(
        "iau03n.dat",
        columns_real=range(6, 14),
        columns_int=range(0, 5),
        conv_factor=CONVRTM,
    )
    iau06arr.appl, iau06arr.appli = load_data(
        "iau03pl.dat",
        columns_real=range(16, 21),  # include column 21 (extra)
        columns_int=range(1, 15),
        conv_factor=CONVRTM,
        convert_exclude_last=True,
    )

    return iau06arr


def iau06in(data_dir: str = DATA_DIR) -> IAU06Array:
    """Initializes the matrices needed for IAU 2006 reduction calculations.

    Args:
        data_dir (str, optional): Path to the directory containing the IAU 2006 data
                                  files (default: DATA_DIR)

    Returns:
        IAU06Array: Dataclass containing all coefficients for IAU 2006 calculations

    Notes:
        Data files are from the IAU 2006 precession-nutation model:
            - iau06xtab5.2.a.dat (file for X coefficients)
            - iau06ytab5.2.b.dat (file for Y coefficients)
            - iau06stab5.2.d.dat (file for S coefficients)
            - iau06gsttab5.2.e.dat (file for GST coefficients)
            - iau06ansofa.dat (file for SOFA luni-solar nutation coefficients)
            - iau06anplsofa.dat (file for SOFA planetary nutation coefficients)
            - iau06nlontab5.3.a.dat (file for nutation in obliquity coefficients
    """

    def load_data(filename, cols_real, cols_int, conv_factor):
        """Helper function to load and process data."""
        filepath = os.path.join(data_dir, filename)
        data = np.loadtxt(filepath, skiprows=2)
        reals = data[:, cols_real[0] : cols_real[1]]
        if conv_factor:
            reals *= conv_factor
        integers = data[:, cols_int[0] : cols_int[1]].astype(int)
        return reals, integers

    # Initialize data object
    iau06arr = IAU06Array()

    # Load all data files
    iau06arr.ax0, iau06arr.ax0i = load_data(
        "iau06xtab5.2.a.dat", cols_real=(1, 3), cols_int=(3, 17), conv_factor=CONVRTU
    )
    iau06arr.ay0, iau06arr.ay0i = load_data(
        "iau06ytab5.2.b.dat", cols_real=(1, 3), cols_int=(3, 17), conv_factor=CONVRTU
    )
    iau06arr.as0, iau06arr.as0i = load_data(
        "iau06stab5.2.d.dat", cols_real=(1, 3), cols_int=(3, 17), conv_factor=CONVRTU
    )
    iau06arr.agst, iau06arr.agsti = load_data(
        "iau06gsttab5.2.e.dat", cols_real=(1, 3), cols_int=(3, 17), conv_factor=CONVRTU
    )
    iau06arr.aapn0, iau06arr.aapn0i = load_data(
        "iau06ansofa.dat", cols_real=(5, 11), cols_int=(0, 5), conv_factor=CONVRTM
    )
    iau06arr.apl0, iau06arr.apl0i = load_data(
        "iau06anplsofa.dat", cols_real=(14, 18), cols_int=(0, 14), conv_factor=CONVRTM
    )
    iau06arr.apn0, iau06arr.apn0i = load_data(
        "iau06nlontab5.3.a.dat", cols_real=(1, 3), cols_int=(3, 17), conv_factor=CONVRTM
    )

    return iau06arr


def readxys(data_dir: str = DATA_DIR) -> IAU06xysArray:
    """Initializes the XYS IAU2006 data from the input file into a dataclass.

    Args:
        data_dir (str, optional): Directory containing the XYS data file "xysdata.dat"
                                  (default: DATA_DIR)

    Returns:
        IAU06xysArray: Dataclass containing the XYS data
    """
    # Read the file as an np.ndarray
    data = np.loadtxt(os.path.join(data_dir, "xysdata.dat"))

    # Create instance and extract individual columns
    iau06xysarr = IAU06xysArray()
    iau06xysarr.jd = data[:, 0]
    iau06xysarr.jdf = data[:, 1]
    iau06xysarr.x = data[:, 2]
    iau06xysarr.y = data[:, 3]
    iau06xysarr.s = data[:, 4]

    # Compute the derived column 'mjd_tt'
    iau06xysarr.mjd_tt = iau06xysarr.jd + iau06xysarr.jdf - JD_TO_MJD_OFFSET

    return iau06xysarr


def readeop(filepath: str) -> EOPArray:
    """Reads the EOP coefficients from a file into an EOPArray dataclass.

    Args:
        filepath (str): Path to the EOP file

    Returns:
        EOPArray: Dataclass containing the EOP values
    """
    # Initialize lists to hold the data
    mjd, xp, yp, dut1, lod, ddpsi, ddeps, dx, dy, dat = ([] for _ in range(10))
    i = 0

    # Read in EOP file
    with open(filepath) as infile:
        lines = infile.readlines()

    # Process each line in the file
    while i < len(lines):
        line = lines[i].strip()

        # Process NUM_OBSERVED_POINTS and NUM_PREDICTED_POINTS
        if "NUM_OBSERVED_POINTS" in line or "NUM_PREDICTED_POINTS" in line:
            num_records = int(line.split()[-1])  # extract number of records
            i += 2  # move to the next line containing data

            for _ in range(num_records):
                if i >= len(lines):
                    break

                # Extract data from the line
                data_line = lines[i].strip()

                try:
                    # Adjust fixed-width indices based on file structure
                    mjd.append(float(data_line[10:16]))
                    xp.append(float(data_line[16:26]))
                    yp.append(float(data_line[26:36]))
                    dut1.append(float(data_line[36:47]))
                    lod.append(float(data_line[47:58]))
                    ddpsi.append(float(data_line[58:68]))
                    ddeps.append(float(data_line[68:78]))
                    dx.append(float(data_line[78:88]))
                    dy.append(float(data_line[88:98]))
                    dat.append(float(data_line[98:102]))
                except ValueError:
                    # Skip lines that do not match the expected format
                    logger.warning(f"Skipping malformed line: {data_line}")
                i += 1

        # Move to the next line if no match
        else:
            i += 1

    return EOPArray(
        mjd=np.array(mjd),
        xp=np.array(xp),
        yp=np.array(yp),
        dut1=np.array(dut1),
        lod=np.array(lod),
        ddpsi=np.array(ddpsi),
        ddeps=np.array(ddeps),
        dx=np.array(dx),
        dy=np.array(dy),
        dat=np.array(dat, dtype=int),
    )


def readspw(filepath: str) -> SPWArray:
    """Reads the Space Weather data from a file into an SPWArray dataclass.

    Args:
        filepath (str): Path to the Space Weather file

    Returns:
        SPWArray: Dataclass containing the Space Weather data

    Notes:
        - The data is first indexed and then converted to values for backwards
          compatibility with older files that contained missing data.
    """
    # Initialize lists to hold the data
    mjd, kparray, sumkp, aparray, avgap = [], [], [], [], []
    adjf107, adjctrf81, adjlstf81 = [], [], []
    obsf107, obsctrf81, obslstf81 = [], [], []
    i = 0

    # Read in Space Weather file
    with open(filepath) as file:
        lines = file.readlines()

    # Process each line in the file
    while i < len(lines):
        line = lines[i].strip()

        # Process NUM_OBSERVED_POINTS and NUM_DAILY_PREDICTED_POINTS
        if "NUM_OBSERVED_POINTS" in line or "NUM_DAILY_PREDICTED_POINTS" in line:
            num_records = int(line.split()[-1])  # extract number of records
            i += 2  # move to the next line containing data

            for _ in range(num_records):
                if i >= len(lines):
                    break

                # Extract data from the line
                data_line = lines[i]

                try:
                    # Extract values using fixed-width indices
                    year = int(data_line[0:4].strip())
                    month = int(data_line[5:8].strip())
                    day = int(data_line[8:11].strip())
                    kparray.append(
                        [float(data_line[i : i + 3].strip()) for i in range(19, 43, 3)]
                    )
                    sumkp.append(float(data_line[43:46].strip()))
                    aparray.append(
                        [float(data_line[i : i + 4].strip()) for i in range(47, 79, 4)]
                    )
                    avgap.append(float(data_line[80:82].strip()))
                    adjf107.append(float(data_line[93:98].strip()))
                    adjctrf81.append(float(data_line[101:107].strip()))
                    adjlstf81.append(float(data_line[107:113].strip()))
                    obsf107.append(float(data_line[113:119].strip()))
                    obsctrf81.append(float(data_line[119:125].strip()))
                    obslstf81.append(float(data_line[125:131].strip()))

                    # Convert date to MJD
                    jd, jdf = jday(year, month, day)
                    mjd.append(jd + jdf - JD_TO_MJD_OFFSET)

                except ValueError:
                    # Skip lines that do not match the expected format
                    logger.warning(f"Skipping malformed line: {data_line}")

                i += 1

        # Move to the next line if no match
        else:
            i += 1

    # Convert lists to numpy arrays
    return SPWArray(
        kparray=np.array(kparray),
        sumkp=np.array(sumkp),
        aparray=np.array(aparray),
        avgap=np.array(avgap),
        adjf107=np.array(adjf107),
        adjctrf81=np.array(adjctrf81),
        adjlstf81=np.array(adjlstf81),
        obsf107=np.array(obsf107),
        obsctrf81=np.array(obsctrf81),
        obslstf81=np.array(obslstf81),
        mjd=np.array(mjd),
    )



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 1 March 2001
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
from typing import Tuple

import numpy as np

from ... import constants as const
from ..time.sidereal import lstime
from .utils import sun_ecliptic_parameters, obliquity_ecliptic


# Set up logging
logger = logging.getLogger(__name__)


def get_moon_latlon(ttdb: float) -> Tuple[float, float]:
    """Calculates the ecliptic/mean longitude of the Moon.

    Args:
        ttdb (float): Julian centuries from J2000

    Returns:
        tuple: (lon, lat)
            lon (float): Ecliptic longitude of the Moon in radians
            lat (float): Ecliptic latitude of the Moon in radians
    """
    lon = (
        float(
            np.radians(
                218.32
                + 481267.8813 * ttdb
                + 6.29 * np.sin(np.radians(134.9 + 477198.85 * ttdb))
                - 1.27 * np.sin(np.radians(259.2 - 413335.38 * ttdb))
                + 0.66 * np.sin(np.radians(235.7 + 890534.23 * ttdb))
                + 0.21 * np.sin(np.radians(269.9 + 954397.7 * ttdb))
                - 0.19 * np.sin(np.radians(357.5 + 35999.05 * ttdb))
                - 0.11 * np.sin(np.radians(186.6 + 966404.05 * ttdb))
            )
        )
        % const.TWOPI
    )

    lat = (
        float(
            np.radians(
                5.13 * np.sin(np.radians(93.3 + 483202.03 * ttdb))
                + 0.28 * np.sin(np.radians(228.2 + 960400.87 * ttdb))
                - 0.28 * np.sin(np.radians(318.3 + 6003.18 * ttdb))
                - 0.17 * np.sin(np.radians(217.6 - 407332.2 * ttdb))
            )
        )
        % const.TWOPI
    )

    return lon, lat


def get_geodetic_dir_cosines(ttdb: float) -> Tuple[float, float, float]:
    """Calculates the geocentric direction cosines of the Moon.

    Args:
        ttdb (float): Julian centuries from J2000

    Returns:
        tuple: (l, m, n)
            l (float): Geocentric direction cosine
            m (float): Geocentric direction cosine
            n (float): Geocentric direction cosine
    """
    # Ecliptic longitude and latitude in radians
    eclplong, eclplat = get_moon_latlon(ttdb)

    # Obliquity of the ecliptic in radians
    obliquity = obliquity_ecliptic(ttdb)

    # Geocentric direction cosines
    l = np.cos(eclplat) * np.cos(eclplong)  # noqa: E741
    m = np.cos(obliquity) * np.cos(eclplat) * np.sin(eclplong) - np.sin(
        obliquity
    ) * np.sin(eclplat)
    n = np.sin(obliquity) * np.cos(eclplat) * np.sin(eclplong) + np.cos(
        obliquity
    ) * np.sin(eclplat)

    return l, m, n


def position(jd: float) -> Tuple[np.ndarray, float, float]:
    """Calculates the geocentric equatorial position vector of the Moon.

    References:
        Vallado: 2022, p. 294, Algorithm 31

    Args:
        jd (float): Julian date (days from 4713 BC)

    Returns:
        tuple: (rmoon, rtasc, decl)
            rmoon (np.ndarray): Inertial moon position vector in km
            rtasc (float): Right ascension of the moon in radians
            decl (float): Declination of the moon in radians
    """
    # Julian centuries from J2000
    ttdb = (jd - const.J2000) / const.CENT2DAY

    # Horizontal parallax (radians)
    hzparal = (
        np.radians(
            0.9508
            + 0.0518 * np.cos(np.radians(134.9 + 477198.85 * ttdb))
            + 0.0095 * np.cos(np.radians(259.2 - 413335.38 * ttdb))
            + 0.0078 * np.cos(np.radians(235.7 + 890534.23 * ttdb))
            + 0.0028 * np.cos(np.radians(269.9 + 954397.7 * ttdb))
        )
        % const.TWOPI
    )

    # Geocentric direction cosines
    l, m, n = get_geodetic_dir_cosines(ttdb)

    # Moon's position vector
    magr = 1 / np.sin(hzparal)
    rmoon = np.array([magr * l, magr * m, magr * n])

    # Right ascension and declination
    rtasc = np.arctan2(m, l)
    decl = np.arcsin(n)

    return rmoon * const.RE, rtasc, decl


def rise_set(
    jd: float, latgd: float, lon: float, n_iters: int = 5, tol: float = 0.008
) -> Tuple[float, float, float]:
    """Finds the universal time for moonrise and moonset given the day and site
    location.

    References:
        Vallado: 2022, p. 296-298, Algorithm 32

    Args:
        jd (float): Julian date (days from 4713 BC)
        latgd (float): Geodetic latitude of the site in radians (-65 deg to 65 deg)
        lon (float): Longitude of the site in radians (-2pi to 2pi) (west is negative)
        n_iters (int, optional): Number of iterations to attempt to find the moon
                                 rise/set times (defaults to 5)
        tol (float, optional): Tolerance for the iteration (defaults to 0.008)

    Returns:
        tuple: (moonrise, moonset, moonphaseang)
            moonrise (float): Universal time of moonrise in hours
            moonset (float): Universal time of moonset in hours
            moonphaseang (float): Moon phase angle in radians
    """
    # Normalize longitude to -π to π
    lon = (lon + np.pi) % const.TWOPI - np.pi

    # Initialize results and variables
    results = {"moonrise": np.nan, "moonset": np.nan}
    moongha, deltaut, ttdb, lha = 0, 0, 0, 0

    # Iteration parameters
    try1 = 1

    for event, jd_offset in [("moonrise", 6), ("moonset", 18)]:
        # Initial guess for UT
        sign = -1 if event == "moonrise" else 1
        uttemp = (jd_offset + sign * np.degrees(lon) / const.DEG2HR) / const.DAY2HR

        # Set if there's a problem
        if try1 == 2:
            uttemp = 0.5

        # Initialize variables for iteration
        i = 0
        tn = uttemp
        t = tn + 10
        jdtemp = jd + uttemp

        while abs(tn - t) >= tol and i <= n_iters:
            # Update the Julian date
            ttdb = (jdtemp - const.J2000) / const.CENT2DAY

            # Ecliptic longitude and latitude in radians
            eclplong, _ = get_moon_latlon(ttdb)

            # Geocentric direction cosines
            l, m, n = get_geodetic_dir_cosines(ttdb)

            # Right ascension and declination
            rtasc = np.arctan2(m, l)
            decl = np.arcsin(n)

            # Correction for right ascension
            if abs(eclplong - rtasc) > np.pi * 0.5:
                rtasc += 0.5 * np.pi * round(0.5 + (eclplong - rtasc) / (0.5 * np.pi))

            # Local sidereal time
            lst, _ = lstime(lon, jdtemp)

            # Calculat hour angles
            moonghan = lst - lon - rtasc
            if i == 0:
                lha = moonghan + lon
                dgha = np.radians(347.81)
            else:
                dgha = (moonghan - moongha) / deltaut
            dgha = dgha + const.TWOPI / abs(deltaut) if dgha < 0 else dgha

            # Local hour angle at moonrise and moonset
            lhan = (0.00233 - np.sin(latgd) * np.sin(decl)) / (
                np.cos(latgd) * np.cos(decl)
            )
            lhan = np.clip(lhan, -1.0, 1.0)
            lhan = np.arccos(lhan)
            if event == "moonrise":
                lhan = const.TWOPI - lhan

            # Time adjustment
            if abs(dgha) > 1e-4:
                deltaut = (lhan - lha) / dgha
            else:
                deltaut = 1.0
                logger.warning("dgha is too small; setting deltaut to 1.0")

            # Adjust deltaut for convergence and handle wrap-around cases
            t = tn
            if abs(deltaut) > 0.5:
                if abs(dgha) > 1e-3:
                    deltaut += (
                        const.TWOPI / dgha if deltaut < 0 else -const.TWOPI / dgha
                    )
                    if abs(deltaut) > 0.51:
                        break

            # Update variables
            tn = uttemp + deltaut
            jdtemp = jdtemp - uttemp + tn
            moongha = moonghan
            i += 1

        # Convert UT to hours
        uttemp = tn * const.DAY2HR if i <= n_iters else 1e4
        uttemp = uttemp % const.DAY2HR if 0.0 <= uttemp < 1e4 else uttemp

        # Assign to results
        results[event] = uttemp

        # Update the iteration and check for solution
        try1 += 1
        if i > n_iters and try1 < 3:  # retry if the first attempt fails
            logger.debug(f"Retrying option {event} (attempt {try1})")
        else:
            if i > n_iters and try1 > 2:  # if all retries fail, set the error
                logger.error(f"Failed to find {event} time")
                results[event] = np.inf
            try1 = 1

    # Mean longitude of the Moon in radians
    meanlonmoon, _ = get_moon_latlon(ttdb)

    # Ecliptic longitude of the Sun in radians
    *_, loneclsun = sun_ecliptic_parameters(ttdb)

    # Moon phase angle
    moonphaseang = (meanlonmoon - loneclsun) % const.TWOPI

    return results["moonrise"], results["moonset"], moonphaseang


def illumination(f: float, elev: float) -> float:
    """Calculates the illumination due to the moon.

    References:
        Vallado: 2022, p. 316-317, Eq. 5-10, Table 5-1

    Args:
        f (float): Phase angle in radians
        elev (float): Moon elevation in radians

    Returns:
        float: Luminous emmittance, lux (lumen/m²)
    """
    # Convert inputs to degrees
    f = np.degrees(f)
    elev = np.degrees(elev)

    # Determine coefficients based on moon elevation
    # See Table 5-1 in Vallado 2022 (p. 317)
    if elev >= 20:
        l0, l1, l2, l3 = -1.95, 4.06, -4.24, 1.56
    elif 5 <= elev < 20:
        l0, l1, l2, l3 = -2.58, 12.58, -42.58, 59.06
    elif -0.8 < elev < 5:
        l0, l1, l2, l3 = -2.79, 24.27, -252.95, 1321.29
    else:
        l0, l1, l2, l3 = 0, 0, 0, 0
        f = 0

    # Compute l1 and l2 based on coefficients and phase angle
    x = elev / 90
    l1 = l0 + l1 * x + l2 * x**2 + l3 * x**3
    l2 = -0.00868 * f - 2.2e-9 * f**4

    # Compute moon illumination
    moonillum = 10 ** (l1 + l2)

    # Clamp moonillum to valid range
    if moonillum < 0 or moonillum >= 1:
        moonillum = 0

    return moonillum



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 31 Oct 2003
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
from enum import Enum
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ... import constants as const


# Set up logging
logger = logging.getLogger(__name__)


class EarthModel(Enum):
    SPHERICAL = "s"
    ELLIPSOIDAL = "e"


def in_sight(
    r1: ArrayLike, r2: ArrayLike, earth_model: EarthModel = EarthModel.ELLIPSOIDAL
) -> bool:
    """Determines if there is line-of-sight (LOS) between two satellites, considering
    the Earth's shape.

    References:
        Vallado: 2022, p. 312-315, Algorithm 35

    Args:
        r1 (array_like): Position vector of the first satellite in km
        r2 (array_like): Position vector of the second satellite in km
        earth_model (EarthModel, optional): Earth model to use (default is ELLIPSOIDAL)

    Returns:
        bool: True if there is line-of-sight, False otherwise
    """
    # Magnitudes
    magr1 = np.linalg.norm(r1)
    magr2 = np.linalg.norm(r2)

    # Scale z-components for ellipsoidal Earth
    temp = (
        1 / np.sqrt(1 - const.ECCEARTHSQRD)
        if earth_model == EarthModel.ELLIPSOIDAL
        else 1
    )
    tr1 = np.array([r1[0], r1[1], r1[2] * temp])
    tr2 = np.array([r2[0], r2[1], r2[2] * temp])

    # Compute magnitudes and dot product
    asqrd = magr1**2
    bsqrd = magr2**2
    adotb = np.dot(tr1, tr2)

    # Compute minimum parametric value
    if abs(asqrd + bsqrd - 2 * adotb) < 1e-4:
        tmin = 0
    else:
        tmin = (asqrd - adotb) / (asqrd + bsqrd - 2 * adotb)
    logger.debug(f"Minimum parametric value (tmin): {tmin}")

    # Check line-of-sight (LOS)
    if tmin < 0 or tmin > 1:
        return True
    else:
        distsqrd = ((1 - tmin) * asqrd + adotb * tmin) / const.RE**2
        return True if distsqrd > 1 else False


def sun_ecliptic_parameters(t: float) -> Tuple[float, float, float]:
    """Compute the mean longitude, mean anomaly, and ecliptic longitude of the Sun.

    Args:
        t (float): Time since J2000 in Julian centuries (e.g. 'tut1' or 'ttdb')

    Returns:
        tuple: (mean_lon, mean_anomaly, ecliptic_lon)
            mean_lon (float): Mean longitude of the Sun in radians
            mean_anomaly (float): Mean anomaly of the Sun in radians
            ecliptic_lon (float): Ecliptic longitude of the Sun in radians
    """
    mean_lon = np.radians(280.4606184 + 36000.77005361 * t) % const.TWOPI
    mean_anomaly = np.radians(357.5277233 + 35999.05034 * t) % const.TWOPI
    ecliptic_lon = (
        np.radians(
            np.degrees(mean_lon)
            + 1.914666471 * np.sin(mean_anomaly)
            + 0.019994643 * np.sin(2 * mean_anomaly)
        )
        % const.TWOPI
    )

    return float(mean_lon), float(mean_anomaly), ecliptic_lon


def obliquity_ecliptic(t: float) -> float:
    """Compute the obliquity of the ecliptic.

    Args:
        t (float): Time since J2000 in Julian centuries (e.g. 'tut1' or 'ttdb')

    Returns:
        float: Obliquity of the ecliptic in radians
    """
    return float(np.radians(np.degrees(const.OBLIQUITYEARTH) - 0.0130042 * t))



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 10 Oct 2019
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from dataclasses import dataclass

import numpy as np


@dataclass
class GravityFieldData:
    c: np.ndarray = None
    s: np.ndarray = None
    c_unc: np.ndarray = None
    s_unc: np.ndarray = None
    normalized: bool = False


def read_gravity_field(filename: str, normalized: bool) -> GravityFieldData:
    """Reads and stores gravity field coefficients.

    References:
        Vallado: 2022, p. 550-551

    Args:
        filename (str): The filename of the gravity field data
        normalized (bool): True if the gravity field data is normalized

    Returns:
        GravityFieldData: A dataclass containing gravity field data:
            - c (np.ndarray): Cosine coefficients
            - s (np.ndarray): Sine coefficients
            - normalized (bool): True if the gravity field data is normalized
    """
    # Load gravity field data
    file_data = np.loadtxt(filename)

    # Get the maximum degree of the gravity field
    max_degree = int(np.max(file_data[:, 0]))
    size = max_degree + 1

    # Initialize gravity field data
    gravarr = GravityFieldData(
        c=np.zeros((size, size)), s=np.zeros((size, size)), normalized=normalized
    )

    # Check if uncertainties are included in the data (columns 5 and 6)
    has_uncertainty = file_data.shape[1] >= 6
    if has_uncertainty:
        gravarr.c_unc = np.zeros((size, size))
        gravarr.s_unc = np.zeros((size, size))

    # Store gravity field coefficients
    for row in file_data:
        n, m = int(row[0]), int(row[1])
        c_value, s_value = row[2], row[3]
        gravarr.c[n, m] = c_value
        gravarr.s[n, m] = s_value

        if has_uncertainty:
            gravarr.c_unc[n, m] = row[4]
            gravarr.s_unc[n, m] = row[5]

    return gravarr



# --------------------------------------------------------------------------------------
# Authors: David Vallado, Jeff Beck
# Date: 28 June 2005
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from dataclasses import dataclass
from typing import Tuple

import numpy as np

from ... import constants as const
from .utils import SatRec


@dataclass
class DscomOutput:
    # TODO: Do we need to return all these variables?
    sinim: float = 0.0
    cosim: float = 0.0
    sinomm: float = 0.0
    cosomm: float = 0.0
    snodm: float = 0.0
    cnodm: float = 0.0
    day: float = 0.0
    e3: float = 0.0
    ee2: float = 0.0
    em: float = 0.0
    emsq: float = 0.0
    gam: float = 0.0
    peo: float = 0.0
    pgho: float = 0.0
    pho: float = 0.0
    pinco: float = 0.0
    plo: float = 0.0
    rtemsq: float = 0.0
    se2: float = 0.0
    se3: float = 0.0
    sgh2: float = 0.0
    sgh3: float = 0.0
    sgh4: float = 0.0
    sh2: float = 0.0
    sh3: float = 0.0
    si2: float = 0.0
    si3: float = 0.0
    sl2: float = 0.0
    sl3: float = 0.0
    sl4: float = 0.0
    s1: float = 0.0
    s2: float = 0.0
    s3: float = 0.0
    s4: float = 0.0
    s5: float = 0.0
    s6: float = 0.0
    s7: float = 0.0
    ss1: float = 0.0
    ss2: float = 0.0
    ss3: float = 0.0
    ss4: float = 0.0
    ss5: float = 0.0
    ss6: float = 0.0
    ss7: float = 0.0
    sz1: float = 0.0
    sz2: float = 0.0
    sz3: float = 0.0
    sz11: float = 0.0
    sz12: float = 0.0
    sz13: float = 0.0
    sz21: float = 0.0
    sz22: float = 0.0
    sz23: float = 0.0
    sz31: float = 0.0
    sz32: float = 0.0
    sz33: float = 0.0
    xgh2: float = 0.0
    xgh3: float = 0.0
    xgh4: float = 0.0
    xh2: float = 0.0
    xh3: float = 0.0
    xi2: float = 0.0
    xi3: float = 0.0
    xl2: float = 0.0
    xl3: float = 0.0
    xl4: float = 0.0
    nm: float = 0.0
    z1: float = 0.0
    z2: float = 0.0
    z3: float = 0.0
    z11: float = 0.0
    z12: float = 0.0
    z13: float = 0.0
    z21: float = 0.0
    z22: float = 0.0
    z23: float = 0.0
    z31: float = 0.0
    z32: float = 0.0
    z33: float = 0.0
    zmol: float = 0.0
    zmos: float = 0.0


@dataclass
class DsInitOutput:
    atime: float = 0.0
    em: float = 0.0
    inclm: float = 0.0
    nodem: float = 0.0
    argpm: float = 0.0
    nm: float = 0.0
    mm: float = 0.0
    irez: int = 0
    d2201: float = 0.0
    d2211: float = 0.0
    d3210: float = 0.0
    d3222: float = 0.0
    d4410: float = 0.0
    d4422: float = 0.0
    d5220: float = 0.0
    d5232: float = 0.0
    d5421: float = 0.0
    d5433: float = 0.0
    dedt: float = 0.0
    didt: float = 0.0
    dmdt: float = 0.0
    dndt: float = 0.0
    dnodt: float = 0.0
    domdt: float = 0.0
    del1: float = 0.0
    del2: float = 0.0
    del3: float = 0.0
    xfact: float = 0.0
    xlamo: float = 0.0
    xli: float = 0.0
    xni: float = 0.0


class DeepSpace:
    """Class to compute deep space common terms for SGP4.

    References:
        - Hoots, Roehrich, NORAD SpaceTrack Report #3, 1980
        - Hoots, Roehrich, NORAD SpaceTrack Report #6, 1986
        - Hoots, Schumacher, and Glover, 2004
        - Vallado, Crawford, Hujsak, Kelso, 2006

    Args:
        epoch (float): Epoch in days from 0 Jan 1950 0:00:00 UTC
        ep (float): Eccentricity
        inclp (float): Inclination in radians
        nodep (float): Right ascension of ascending node in radians
        argpp (float): Argument of perigee in radians
        np_ (float): Mean motion in radians per minute
        mp (float, optional): Mean anomaly in radians (default = 0)
        use_afspc_mode (bool, optional): Use AFSPC mode for nodep (default = True)

    Attributes:
        epoch (float): Epoch in days from 0 Jan 1950 0:00:00 UTC
        ep (float): Eccentricity
        inclp (float): Inclination in radians
        nodep (float): Right ascension of ascending node in radians
        argpp (float): Argument of perigee in radians
        np_ (float): Mean motion in radians per minute
        mp (float): Mean anomaly in radians
        use_afspc_mode (bool): Use AFSPC mode for nodep
        dscom_out (DscomOutput): Deep space common terms output
        dsinit_out (DsInitOutput): Deep space initialization output

    TODO:
        - This class could be further refactored/cleaned up for better readability.
    """

    def __init__(
        self,
        epoch: float,
        ep: float,
        inclp: float,
        nodep: float,
        argpp: float,
        np_: float,
        mp: float = 0,
        use_afspc_mode: bool = True,
    ):
        self.epoch = epoch
        self.ep = ep
        self.inclp = inclp
        self.nodep = nodep
        self.argpp = argpp
        self.np_ = np_
        self.mp = mp
        self.use_afspc_mode = use_afspc_mode

        # Initialize output dataclasses
        self.dscom_out = None
        self.dsinit_out = None

    def set_attributes(self, **kwargs):
        """Set multiple attributes dynamically."""
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_attributes(self, *args):
        """Get multiple attributes dynamically."""
        return tuple(getattr(self, key) for key in args)

    def set_mean_elems(self, **kwargs):
        """Set mean elements in dsinit_out."""
        if not self.dsinit_out:
            raise ValueError("dsinit_out not set. Run dsinit() first.")
        for key, value in kwargs.items():
            if hasattr(self.dsinit_out, key):
                setattr(self.dsinit_out, key, value)
            else:
                raise AttributeError(f"'DsInitOutput' object has no attribute '{key}'")

    def get_mean_elems(self) -> Tuple[float, float, float, float, float, float]:
        """Get the dsinit_out mean elements."""
        if not self.dsinit_out:
            raise ValueError("dsinit_out not set. Run dsinit() first.")
        return (
            self.dsinit_out.em,
            self.dsinit_out.inclm,
            self.dsinit_out.nodem,
            self.dsinit_out.argpm,
            self.dsinit_out.nm,
            self.dsinit_out.mm,
        )

    def dscom(self, tc: float):
        """Computes deep space common terms for SGP4 (used by both the secular and
        periodics subroutines).

        Args:
            tc (float): Time correction in minutes

        Returns:
            None (sets dscom_out attribute)
        """
        # Initialize output dataclass
        out = DscomOutput()

        # Constants
        zes, zel = 0.01675, 0.0549
        c1ss, c1l = 2.9864797e-6, 4.7968065e-7
        zsinis, zcosis = 0.39785416, 0.91744867
        zcosgs, zsings = 0.1945905, -0.98088458

        # Initialize variables
        out.nm, out.em = self.np_, self.ep
        out.snodm, out.cnodm = np.sin(self.nodep), np.cos(self.nodep)
        out.sinomm, out.cosomm = np.sin(self.argpp), np.cos(self.argpp)
        out.sinim, out.cosim = np.sin(self.inclp), np.cos(self.inclp)
        out.emsq = out.em**2
        betasq = 1 - out.emsq
        out.rtemsq = np.sqrt(betasq)

        # Initialize lunar solar terms
        out.day = self.epoch + 18261.5 + tc / const.DAY2MIN
        xnodce = np.remainder(4.523602 - 9.2422029e-4 * out.day, const.TWOPI)
        stem, ctem = np.sin(xnodce), np.cos(xnodce)
        zcosil = 0.91375164 - 0.03568096 * ctem
        zsinil = np.sqrt(1 - zcosil**2)
        zsinhl = 0.089683511 * stem / zsinil
        zcoshl = np.sqrt(1 - zsinhl**2)
        out.gam = 5.8351514 + 0.001944368 * out.day
        zy = zcoshl * ctem + 0.91744867 * zsinhl * stem
        zx = np.arctan2(0.39785416 * stem / zsinil, zy)
        zx = out.gam + zx - xnodce
        zcosgl, zsingl = np.cos(zx), np.sin(zx)

        # Solar and lunar terms
        zcosg, zsing = zcosgs, zsings
        zcosi, zsini = zcosis, zsinis
        zcosh, zsinh = out.cnodm, out.snodm
        cc = c1ss

        # Loop for solar and lunar terms
        for lsflg in range(2):
            # Compute intermediate values
            a1 = zcosg * zcosh + zsing * zcosi * zsinh
            a3 = -zsing * zcosh + zcosg * zcosi * zsinh
            a7 = -zcosg * zsinh + zsing * zcosi * zcosh
            a8 = zsing * zsini
            a9 = zsing * zsinh + zcosg * zcosi * zcosh
            a10 = zcosg * zsini
            a2 = out.cosim * a7 + out.sinim * a8
            a4 = out.cosim * a9 + out.sinim * a10
            a5 = -out.sinim * a7 + out.cosim * a8
            a6 = -out.sinim * a9 + out.cosim * a10

            # Further calculations
            x1 = a1 * out.cosomm + a2 * out.sinomm
            x2 = a3 * out.cosomm + a4 * out.sinomm
            x3 = -a1 * out.sinomm + a2 * out.cosomm
            x4 = -a3 * out.sinomm + a4 * out.cosomm
            x5 = a5 * out.sinomm
            x6 = a6 * out.sinomm
            x7 = a5 * out.cosomm
            x8 = a6 * out.cosomm

            # Z matrix calculations
            out.z31 = 12 * x1**2 - 3 * x3**2
            out.z32 = 24 * x1 * x2 - 6 * x3 * x4
            out.z33 = 12 * x2**2 - 3 * x4**2
            out.z1 = 3 * (a1**2 + a2**2) + out.z31 * out.emsq
            out.z2 = 6 * (a1 * a3 + a2 * a4) + out.z32 * out.emsq
            out.z3 = 3 * (a3**2 + a4**2) + out.z33 * out.emsq
            out.z11 = -6 * a1 * a5 + out.emsq * (-24 * x1 * x7 - 6 * x3 * x5)
            out.z12 = -6 * (a1 * a6 + a3 * a5) + out.emsq * (
                -24 * (x2 * x7 + x1 * x8) - 6 * (x3 * x6 + x4 * x5)
            )
            out.z13 = -6 * a3 * a6 + out.emsq * (-24 * x2 * x8 - 6 * x4 * x6)
            out.z21 = 6 * a2 * a5 + out.emsq * (24 * x1 * x5 - 6 * x3 * x7)
            out.z22 = 6 * (a4 * a5 + a2 * a6) + out.emsq * (
                24 * (x2 * x5 + x1 * x6) - 6 * (x4 * x7 + x3 * x8)
            )
            out.z23 = 6 * a4 * a6 + out.emsq * (24 * x2 * x6 - 6 * x4 * x8)

            # Update z values with beta square terms
            out.z1 += out.z1 + betasq * out.z31
            out.z2 += out.z2 + betasq * out.z32
            out.z3 += out.z3 + betasq * out.z33

            # S coefficients
            out.s3 = cc * 1 / out.nm
            out.s2 = -0.5 * out.s3 / out.rtemsq
            out.s4 = out.s3 * out.rtemsq
            out.s1 = -15 * out.em * out.s4
            out.s5 = x1 * x3 + x2 * x4
            out.s6 = x2 * x3 + x1 * x4
            out.s7 = x2 * x4 - x1 * x3

            # Update lunar terms on first iteration
            if lsflg == 0:
                out.ss1 = out.s1
                out.ss2 = out.s2
                out.ss3 = out.s3
                out.ss4 = out.s4
                out.ss5 = out.s5
                out.ss6 = out.s6
                out.ss7 = out.s7
                out.sz1 = out.z1
                out.sz2 = out.z2
                out.sz3 = out.z3
                out.sz11 = out.z11
                out.sz12 = out.z12
                out.sz13 = out.z13
                out.sz21 = out.z21
                out.sz22 = out.z22
                out.sz23 = out.z23
                out.sz31 = out.z31
                out.sz32 = out.z32
                out.sz33 = out.z33
                zcosg, zsing = zcosgl, zsingl
                zcosi, zsini = zcosil, zsinil
                zcosh = zcoshl * out.cnodm + zsinhl * out.snodm
                zsinh = out.snodm * zcoshl - out.cnodm * zsinhl
                cc = c1l

        # Compute additional terms
        out.zmol = np.remainder(4.7199672 + 0.2299715 * out.day - out.gam, const.TWOPI)
        out.zmos = np.remainder(6.2565837 + 0.017201977 * out.day, const.TWOPI)

        # Compute solar terms
        out.se2 = 2 * out.ss1 * out.ss6
        out.se3 = 2 * out.ss1 * out.ss7
        out.si2 = 2 * out.ss2 * out.sz12
        out.si3 = 2 * out.ss2 * (out.sz13 - out.sz11)
        out.sl2 = -2 * out.ss3 * out.sz2
        out.sl3 = -2 * out.ss3 * (out.sz3 - out.sz1)
        out.sl4 = -2 * out.ss3 * (-21 - 9 * out.emsq) * zes
        out.sgh2 = 2 * out.ss4 * out.sz32
        out.sgh3 = 2 * out.ss4 * (out.sz33 - out.sz31)
        out.sgh4 = -18 * out.ss4 * zes
        out.sh2 = -2 * out.ss2 * out.sz22
        out.sh3 = -2 * out.ss2 * (out.sz23 - out.sz21)

        # Compute lunar terms
        out.ee2 = 2 * out.s1 * out.s6
        out.e3 = 2 * out.s1 * out.s7
        out.xi2 = 2 * out.s2 * out.z12
        out.xi3 = 2 * out.s2 * (out.z13 - out.z11)
        out.xl2 = -2 * out.s3 * out.z2
        out.xl3 = -2 * out.s3 * (out.z3 - out.z1)
        out.xl4 = -2 * out.s3 * (-21 - 9 * out.emsq) * zel
        out.xgh2 = 2 * out.s4 * out.z32
        out.xgh3 = 2 * out.s4 * (out.z33 - out.z31)
        out.xgh4 = -18 * out.s4 * zel
        out.xh2 = -2 * out.s2 * out.z22
        out.xh3 = -2 * out.s2 * (out.z23 - out.z21)

        self.dscom_out = out

    def dpper(self, t: float, incl_tol: float = 0.2):
        """Deep space long period periodic contributions to mean elements.

        The method `dscom()` must be called prior to running this method!

        Args:
            t (float): Elapsed time since epoch in minutes
            incl_tol (float): Inclination tolerance for periodics (default = 0.2)

        Returns:
            None (updates class attributes)

        Notes:
            - These periodics are zero at epoch by design.
        """
        # Check if dscom_out is set
        if not self.dscom_out:
            raise ValueError("dscom_out not set. Run dscom() first.")

        # Constants
        zns, zes, znl, zel = 1.19459e-5, 0.01675, 1.5835218e-4, 0.0549

        # Extract variables
        ee2, e3 = self.dscom_out.ee2, self.dscom_out.e3
        se2, se3 = self.dscom_out.se2, self.dscom_out.se3
        si2, si3 = self.dscom_out.si2, self.dscom_out.si3
        sl2, sl3, sl4 = self.dscom_out.sl2, self.dscom_out.sl3, self.dscom_out.sl4
        sgh2, sgh3, sgh4 = self.dscom_out.sgh2, self.dscom_out.sgh3, self.dscom_out.sgh4
        sh2, sh3 = self.dscom_out.sh2, self.dscom_out.sh3
        xi2, xi3 = self.dscom_out.xi2, self.dscom_out.xi3
        xl2, xl3, xl4 = self.dscom_out.xl2, self.dscom_out.xl3, self.dscom_out.xl4
        xgh2, xgh3, xgh4 = self.dscom_out.xgh2, self.dscom_out.xgh3, self.dscom_out.xgh4
        xh2, xh3 = self.dscom_out.xh2, self.dscom_out.xh3
        zmol, zmos = self.dscom_out.zmol, self.dscom_out.zmos

        # Calculate time-varying periodics
        zm = zmos + zns * t
        zf = zm + 2 * zes * np.sin(zm)
        sinzf = np.sin(zf)
        f2 = 0.5 * sinzf**2 - 0.25
        f3 = -0.5 * sinzf * np.cos(zf)
        ses = se2 * f2 + se3 * f3
        sis = si2 * f2 + si3 * f3
        sls = sl2 * f2 + sl3 * f3 + sl4 * sinzf
        sghs = sgh2 * f2 + sgh3 * f3 + sgh4 * sinzf
        shs = sh2 * f2 + sh3 * f3
        zm = zmol + znl * t
        zf = zm + 2 * zel * np.sin(zm)
        sinzf = np.sin(zf)
        f2 = 0.5 * sinzf**2 - 0.25
        f3 = -0.5 * sinzf * np.cos(zf)
        sel = ee2 * f2 + e3 * f3
        sil = xi2 * f2 + xi3 * f3
        sll = xl2 * f2 + xl3 * f3 + xl4 * sinzf
        sghl = xgh2 * f2 + xgh3 * f3 + xgh4 * sinzf
        shll = xh2 * f2 + xh3 * f3

        # Initialize periodics
        pe = ses + sel
        pinc = sis + sil
        pl = sls + sll
        pgh = sghs + sghl
        ph = shs + shll

        # Subtract baseline offsets
        pe -= self.dscom_out.peo
        pinc -= self.dscom_out.pinco
        pl -= self.dscom_out.plo
        pgh -= self.dscom_out.pgho
        ph -= self.dscom_out.pho

        # Update inclination and eccentricity
        self.inclp += pinc
        self.ep += pe
        sinip = np.sin(self.inclp)
        cosip = np.cos(self.inclp)

        # Apply periodics based on inclination
        if self.inclp >= incl_tol:
            # Apply periodics directly
            ph /= sinip
            pgh -= cosip * ph
            self.argpp += pgh
            self.nodep += ph
            self.mp += pl
            return

        # Apply periodics with Lyddane modification
        sinop, cosop = np.sin(self.nodep), np.cos(self.nodep)
        alfdp = sinip * sinop
        betdp = sinip * cosop
        dalf = ph * cosop + pinc * cosip * sinop
        dbet = -ph * sinop + pinc * cosip * cosop
        alfdp += dalf
        betdp += dbet

        # SGP4 fix for AFSPC-written intrinsic functions
        self.nodep = np.remainder(self.nodep, const.TWOPI)
        if self.nodep < 0 and self.use_afspc_mode:
            self.nodep += const.TWOPI

        xls = self.mp + self.argpp + cosip * self.nodep
        dls = pl + pgh - pinc * self.nodep * sinip
        xls += dls
        xnoh = self.nodep
        self.nodep = np.arctan2(alfdp, betdp)

        # SGP4 fix for AFSPC-written intrinsic functions
        if self.nodep < 0 and self.use_afspc_mode:
            self.nodep += const.TWOPI

        if np.abs(xnoh - self.nodep) > np.pi:
            self.nodep += const.TWOPI if self.nodep < xnoh else -const.TWOPI

        self.mp += pl
        self.argpp = xls - self.mp - cosip * self.nodep

    def _calc_geopotential_resonance(
        self,
        satrec,
        out,
        aonv,
        eccsq,
        root22,
        root32,
        root44,
        root52,
        root54,
        rptim,
        theta,
    ):
        cosim, sinim = self.dscom_out.cosim, self.dscom_out.sinim
        cosisq = self.dscom_out.cosim**2
        emo, em, emsq = out.em, satrec.ecco, eccsq
        eoc = em * emsq
        g201 = -0.306 - (em - 0.64) * 0.44

        if em <= 0.65:
            g211 = 3.616 - 13.247 * em + 16.29 * emsq
            g310 = -19.302 + 117.39 * em - 228.419 * emsq + 156.591 * eoc
            g322 = -18.9068 + 109.7927 * em - 214.6334 * emsq + 146.5816 * eoc
            g410 = -41.122 + 242.694 * em - 471.094 * emsq + 313.953 * eoc
            g422 = -146.407 + 841.88 * em - 1629.014 * emsq + 1083.435 * eoc
            g520 = -532.114 + 3017.977 * em - 5740.032 * emsq + 3708.276 * eoc
        else:
            g211 = -72.099 + 331.819 * em - 508.738 * emsq + 266.724 * eoc
            g310 = -346.844 + 1582.851 * em - 2415.925 * emsq + 1246.113 * eoc
            g322 = -342.585 + 1554.908 * em - 2366.899 * emsq + 1215.972 * eoc
            g410 = -1052.797 + 4758.686 * em - 7193.992 * emsq + 3651.957 * eoc
            g422 = -3581.69 + 16178.11 * em - 24462.77 * emsq + 12422.52 * eoc
            if em > 0.715:
                g520 = -5149.66 + 29936.92 * em - 54087.36 * emsq + 31324.56 * eoc
            else:
                g520 = 1464.74 - 4664.75 * em + 3763.64 * emsq

        if em < 0.7:
            g533 = -919.2277 + 4988.61 * em - 9064.77 * emsq + 5542.21 * eoc
            g521 = -822.71072 + 4568.6173 * em - 8491.4146 * emsq + 5337.524 * eoc
            g532 = -853.666 + 4690.25 * em - 8624.77 * emsq + 5341.4 * eoc
        else:
            g533 = -37995.78 + 161616.52 * em - 229838.2 * emsq + 109377.94 * eoc
            g521 = -51752.104 + 218913.95 * em - 309468.16 * emsq + 146349.42 * eoc
            g532 = -40023.88 + 170470.89 * em - 242699.48 * emsq + 115605.82 * eoc

        out.em = em
        sini2 = sinim**2
        f220 = 0.75 * (1 + 2 * cosim + cosisq)
        f221 = 1.5 * sini2
        f321 = 1.875 * sinim * (1 - 2 * cosim - 3 * cosisq)
        f322 = -1.875 * sinim * (1 + 2 * cosim - 3 * cosisq)
        f441 = 35 * sini2 * f220
        f442 = 39.375 * sini2**2
        f522 = (
            9.84375
            * sinim
            * (
                sini2 * (1 - 2 * cosim - 5 * cosisq)
                + 0.33333333 * (-2 + 4 * cosim + 6 * cosisq)
            )
        )
        f523 = sinim * (
            4.92187512 * sini2 * (-2 - 4 * cosim + 10 * cosisq)
            + 6.56250012 * (1 + 2 * cosim - 3 * cosisq)
        )
        f542 = (
            29.53125
            * sinim
            * (2 - 8 * cosim + cosisq * (-12 + 8 * cosim + 10 * cosisq))
        )
        f543 = (
            29.53125
            * sinim
            * (-2 - 8 * cosim + cosisq * (12 + 8 * cosim - 10 * cosisq))
        )

        temp1 = 3 * out.nm**2 * aonv**2
        temp = temp1 * root22
        out.d2201 = temp * f220 * g201
        out.d2211 = temp * f221 * g211
        temp1 *= aonv
        temp = temp1 * root32
        out.d3210 = temp * f321 * g310
        out.d3222 = temp * f322 * g322
        temp1 *= aonv
        temp = 2 * temp1 * root44
        out.d4410 = temp * f441 * g410
        out.d4422 = temp * f442 * g422
        temp1 *= aonv
        temp = temp1 * root52
        out.d5220 = temp * f522 * g520
        out.d5232 = temp * f523 * g532
        temp = 2 * temp1 * root54
        out.d5421 = temp * f542 * g521
        out.d5433 = temp * f543 * g533
        out.xlamo = (
            satrec.mo + satrec.nodeo + satrec.nodeo - theta - theta
        ) % const.TWOPI
        out.xfact = (
            satrec.mdot
            + out.dmdt
            + 2 * (satrec.nodedot + out.dnodt - rptim)
            - satrec.no
        )
        out.em = emo

    def _calc_synchronous_resonance(
        self, satrec, out, emsq, q22, q33, q31, aonv, theta, xpidot, rptim
    ):
        g200 = 1 + emsq * (-2.5 + 0.8125 * emsq)
        g310 = 1 + 2 * emsq
        g300 = 1 + emsq * (-6 + 6.60937 * emsq)
        f220 = 0.75 * (1 + self.dscom_out.cosim) ** 2
        f311 = 0.9375 * self.dscom_out.sinim**2 * (
            1 + 3 * self.dscom_out.cosim
        ) - 0.75 * (1 + self.dscom_out.cosim)
        f330 = 1 + self.dscom_out.cosim
        f330 = 1.875 * f330**3
        out.del1 = 3 * out.nm**2 * aonv**2
        out.del2 = 2 * out.del1 * f220 * g200 * q22
        out.del3 = 3 * out.del1 * f330 * g300 * q33 * aonv
        out.del1 *= f311 * g310 * q31 * aonv
        out.xlamo = (satrec.mo + satrec.nodeo + satrec.argpo - theta) % const.TWOPI
        out.xfact = (
            satrec.mdot + xpidot - rptim + out.dmdt + out.domdt + out.dnodt - satrec.no
        )

    def dsinit(
        self,
        satrec: SatRec,
        xke: float,
        tc: float,
        gsto: float,
        xpidot: float,
        eccsq: float,
        inclm: float,
        nodem: float,
        argpm: float,
        mm: float,
        incl_tol: float = np.radians(3),
    ):
        """Deep space initialization for SGP4.

        This function provides deep space contributions to mean motion dot due to
        geopotential resonance with half day and one day orbits.

        The method `dscom()` must be called prior to running this method!

        Args:
            satrec (SatRec): Satellite record dataclass
            xke (float): SGP4 constant
            tc (float): Time correction in minutes
            gsto (float): GST at epoch in radians
            xpidot (float): RAAN dot in rad/s
            eccsq (float): Eccentricity squared
            inclm (float): Inclination in radians
            nodem (float): RAAN in radians
            argpm (float): Argument of perigee in radians
            mm (float): Mean anomaly in radians
            incl_tol (float): Inclination tolerance for applying periodics
                              (default = 3 deg in radians)

        Returns:
            None (sets dsinit_out attribute)
        """
        # Check if dscom_out is set
        if not self.dscom_out:
            raise ValueError("dscom_out not set. Run dscom() first.")

        # Initialize output dataclass
        out = DsInitOutput()

        # Constants
        rptim = 4.37526908801129966e-3
        q22, q31, q33 = 1.7891679e-6, 2.1460748e-6, 2.2123015e-7
        root22, root44, root54 = 1.7891679e-6, 7.3636953e-9, 2.1765803e-9
        root32, root52 = 3.7393792e-7, 1.1428639e-7
        x2o3 = 2 / 3
        znl, zns = 1.5835218e-4, 1.19459e-5

        # Initialize outputs
        out.em, out.nm = self.dscom_out.em, self.dscom_out.nm
        if 0.0034906585 < out.nm < 0.0052359877:
            out.irez = 1
        if 8.26e-3 <= out.nm <= 9.24e-3 and out.em >= 0.5:
            out.irez = 2

        # Extract exising variables
        emsq = self.dscom_out.emsq
        sinim, cosim = self.dscom_out.sinim, self.dscom_out.cosim
        s1, s2, s3 = self.dscom_out.s1, self.dscom_out.s2, self.dscom_out.s3
        s4, s5 = self.dscom_out.s4, self.dscom_out.s5
        ss1, ss2, ss3 = self.dscom_out.ss1, self.dscom_out.ss2, self.dscom_out.ss3
        ss4, ss5 = self.dscom_out.ss4, self.dscom_out.ss5
        sz1, sz3, sz11 = self.dscom_out.sz1, self.dscom_out.sz3, self.dscom_out.sz11
        sz13, sz21, sz23 = self.dscom_out.sz13, self.dscom_out.sz21, self.dscom_out.sz23
        sz31, sz33 = self.dscom_out.sz31, self.dscom_out.sz33
        z1, z3, z11 = self.dscom_out.z1, self.dscom_out.z3, self.dscom_out.z11
        z13, z21, z23 = self.dscom_out.z13, self.dscom_out.z21, self.dscom_out.z23
        z31, z33 = self.dscom_out.z31, self.dscom_out.z33

        # Solar terms
        ses = ss1 * zns * ss5
        sis = ss2 * zns * (sz11 + sz13)
        sls = -zns * ss3 * (sz1 + sz3 - 14 - 6 * emsq)
        sghs = ss4 * zns * (sz31 + sz33 - 6)
        shs = -zns * ss2 * (sz21 + sz23)
        if inclm < incl_tol or inclm > np.pi - incl_tol:
            shs = 0
        if sinim != 0:
            shs /= sinim
        sgs = sghs - cosim * shs

        # Lunar terms
        out.dedt = ses + s1 * znl * s5
        out.didt = sis + s2 * znl * (z11 + z13)
        out.dmdt = sls - znl * s3 * (z1 + z3 - 14 - 6 * emsq)
        sghl = s4 * znl * (z31 + z33 - 6)
        shll = -znl * s2 * (z21 + z23)
        if inclm < incl_tol or inclm > np.pi - incl_tol:
            shll = 0
        out.domdt, out.dnodt = sgs + sghl, shs
        if sinim != 0:
            out.domdt -= cosim / sinim * shll
            out.dnodt += shll / sinim

        # Deep space resonance effects
        theta = np.remainder(gsto + tc * rptim, const.TWOPI)
        out.em += out.dedt * satrec.t
        out.inclm = inclm + out.didt * satrec.t
        out.argpm = argpm + out.domdt * satrec.t
        out.nodem = nodem + out.dnodt * satrec.t
        out.mm = mm + out.dmdt * satrec.t

        # Return if no resonance
        if out.irez == 0:
            return out

        aonv = (out.nm / xke) ** x2o3

        # Geopotential resonance for 12-hour orbits
        if out.irez == 2:
            self._calc_geopotential_resonance(
                satrec,
                out,
                aonv,
                eccsq,
                root22,
                root32,
                root44,
                root52,
                root54,
                rptim,
                theta,
            )

        # Synchronous resonance terms
        if out.irez == 1:
            self._calc_synchronous_resonance(
                satrec, out, emsq, q22, q33, q31, aonv, theta, xpidot, rptim
            )

        # Initialize the integrator for SGP4
        out.xli, out.xni = out.xlamo, satrec.no
        out.nm += out.dndt

        self.dsinit_out = out

    def dspace(self, satrec: SatRec, tc: float, gsto: float):
        """Provides deep space contributions to mean elements for perturbing third body.

        These effects have been averaged over one revolution of the sun and moon. For
        Earth resonance effects, the effects have been averaged over no revolutions of
        the satellite (mean motion).

        The method `dsinit()` must be called prior to running this method!

        Args:
            satrec (SatRec): Satellite record dataclass
            tc (float): Time correction in minutes
            gsto (float): GST at epoch in radians

        Returns:
            None (updates dsinit_out attribute)
        """
        # Check if dsinit_out is set
        if not self.dsinit_out:
            raise ValueError("dsinit_out not set. Run dsinit() first.")

        # Constants
        fasx2, fasx4, fasx6 = 0.13130908, 2.8843198, 0.37448087
        g22, g32, g44, g52, g54 = 5.7686396, 0.95240898, 1.8014998, 1.0508330, 4.4108898
        rptim = 4.37526908801129966e-3
        stepp, stepn, step2 = 720, -720, 259200

        # Extract existing variables
        out = self.dsinit_out
        atime, em, inclm, argpm = out.atime, out.em, out.inclm, out.argpm
        nm, mm, nodem, xli, xni = out.nm, out.mm, out.nodem, out.xli, out.xni
        d2201, d2211, d3210, d3222 = out.d2201, out.d2211, out.d3210, out.d3222
        d4410, d4422, d5220, d5232 = out.d4410, out.d4422, out.d5220, out.d5232
        d5421, d5433 = out.d5421, out.d5433
        dedt, didt, dmdt = out.dedt, out.didt, out.dmdt
        dnodt, domdt = out.dnodt, out.domdt
        irez, xfact, xlamo = out.irez, out.xfact, out.xlamo

        # Initialize variables for deep space resonance effects calculation
        dndt = 0
        theta = np.remainder(gsto + tc * rptim, const.TWOPI)
        em += dedt * satrec.t
        inclm += didt * satrec.t
        argpm += domdt * satrec.t
        nodem += dnodt * satrec.t
        mm += dmdt * satrec.t

        # Return if no resonance
        if irez == 0:
            vars(self.dsinit_out).update(
                {
                    "em": em,
                    "inclm": inclm,
                    "nodem": nodem,
                    "argpm": argpm,
                    "nm": nm,
                    "mm": mm,
                    "dndt": dndt,
                }
            )
            return

        # Update resonances: numerical (Euler-Maclaurin) integration
        # The following integration works for negative time steps and periods
        # The specific changes are unknown because the original code was so convoluted

        # SGP4 fix streamline check
        if atime == 0 or satrec.t * atime <= 0 or abs(satrec.t) < abs(atime):
            atime, xni, xli = 0, satrec.no, xlamo
        delt = stepp if satrec.t >= 0 else stepn

        while True:
            # Near-synchronous resonance terms
            if irez != 2:
                xndt = (
                    out.del1 * np.sin(xli - fasx2)
                    + out.del2 * np.sin(2 * (xli - fasx4))
                    + out.del3 * np.sin(3 * (xli - fasx6))
                )
                xldot = xni + xfact
                xnddt = (
                    out.del1 * np.cos(xli - fasx2)
                    + 2 * out.del2 * np.cos(2 * (xli - fasx4))
                    + 3 * out.del3 * np.cos(3 * (xli - fasx6))
                )
                xnddt *= xldot

            # Near half-day resonance terms
            else:
                xomi = satrec.argpo + satrec.argpdot * atime
                x2omi, x2li = xomi + xomi, xli + xli
                xndt = (
                    d2201 * np.sin(x2omi + xli - g22)
                    + d2211 * np.sin(xli - g22)
                    + d3210 * np.sin(xomi + xli - g32)
                    + d3222 * np.sin(-xomi + xli - g32)
                    + d4410 * np.sin(x2omi + x2li - g44)
                    + d4422 * np.sin(x2li - g44)
                    + d5220 * np.sin(xomi + xli - g52)
                    + d5232 * np.sin(-xomi + xli - g52)
                    + d5421 * np.sin(xomi + x2li - g54)
                    + d5433 * np.sin(-xomi + x2li - g54)
                )
                xldot = xni + xfact
                xnddt = (
                    d2201 * np.cos(x2omi + xli - g22)
                    + d2211 * np.cos(xli - g22)
                    + d3210 * np.cos(xomi + xli - g32)
                    + d3222 * np.cos(-xomi + xli - g32)
                    + d5220 * np.cos(xomi + xli - g52)
                    + d5232 * np.cos(-xomi + xli - g52)
                    + 2
                    * (
                        d4410 * np.cos(x2omi + x2li - g44)
                        + d4422 * np.cos(x2li - g44)
                        + d5421 * np.cos(xomi + x2li - g54)
                        + d5433 * np.cos(-xomi + x2li - g54)
                    )
                )
                xnddt *= xldot

            # Integrator
            if abs(satrec.t - atime) < stepp:
                ft = satrec.t - atime
                break

            xli += xldot * delt + xndt * step2
            xni += xndt * delt + xnddt * step2
            atime += delt

        # Update elements
        nm = xni + xndt * ft + xnddt * ft**2 * 0.5
        xl = xli + xldot * ft + xndt * ft**2 * 0.5
        if irez != 1:
            mm = xl - 2 * nodem + 2 * theta
            dndt = nm - satrec.no
        else:
            mm = xl - nodem - argpm + theta
            dndt = nm - satrec.no
        nm = satrec.no + dndt

        # Update `dsinit_out` with the final computed values
        vars(self.dsinit_out).update(
            {
                "atime": atime,
                "em": em,
                "inclm": inclm,
                "nodem": nodem,
                "argpm": argpm,
                "nm": nm,
                "mm": mm,
                "dndt": dndt,
                "xli": xli,
                "xni": xni,
            }
        )



# --------------------------------------------------------------------------------------
# Authors: Sal Alfano, David Vallado
# Date: 31 March 2011
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike


def poscov2pts(reci: ArrayLike, cov: ArrayLike) -> np.ndarray:
    """Generates 6 sigma points from position and covariance using the Cholesky method.

    Args:
        reci (array_like): 3x1 ECI position vector
        cov (array_like): 3x3 ECI covariance matrix

    Returns:
        np.ndarray: 3x6 matrix of sigma points (position only)

    Notes:
        - Units can be in km or m, but must be consistent
    """
    # Initialize the sigma points matrix
    sigmapts = np.zeros((3, 6))

    # Compute matrix square root using Cholesky decomposition
    s = np.sqrt(3) * np.linalg.cholesky(cov)

    # Generate sigma points
    for i in range(3):
        offset = s[:, i]
        jj = i * 2  # index for positive/negative perturbations

        # Positive perturbation
        sigmapts[:, jj] = np.array(reci) + offset[:3]

        # Negative perturbation
        sigmapts[:, jj + 1] = np.array(reci) - offset[:3]

    return sigmapts


def posvelcov2pts(reci: ArrayLike, veci: ArrayLike, cov: ArrayLike) -> np.ndarray:
    """Generates 12 sigma points from position, velocity, and covariance using the
    Cholesky method.

    Args:
        reci (array_like): 3x1 ECI position vector
        veci (array_like): 3x1 ECI velocity vector
        cov (array_like): 6x6 ECI covariance matrix

    Returns:
        np.ndarray: 6x12 matrix of sigma points (position and velocity)

    Notes:
        - Units can be in km or m, but must be consistent
    """
    # Initialize the sigma points matrix
    sigmapts = np.zeros((6, 12))

    # Compute matrix square root using Cholesky decomposition
    s = np.sqrt(6) * np.linalg.cholesky(cov)

    # Generate sigma points
    for i in range(6):
        offset = s[:, i]
        jj = i * 2  # index for positive/negative perturbations

        # Positive perturbation
        sigmapts[:3, jj] = np.array(reci) + offset[:3]
        sigmapts[3:, jj] = np.array(veci) + offset[3:]

        # Negative perturbation
        sigmapts[:3, jj + 1] = np.array(reci) - offset[:3]
        sigmapts[3:, jj + 1] = np.array(veci) - offset[3:]

    return sigmapts


def remakecov(sigmapts: ArrayLike) -> Tuple[np.ndarray, np.ndarray]:
    """Calculates the mean vector and covariance matrix from propagated sigma points.

    Args:
        sigmapts (array_like): n_dim x n_pts matrix of propagated points from the
        square root algorithm

    Returns:
        tuple: (yu, cov)
            yu (np.ndarray): n_dim x 1 mean vector
            cov (np.ndarray): n_dim x n_dim covariance matrix
    """
    sigmapts = np.array(sigmapts)
    n_dim, n_pts = sigmapts.shape

    # Compute the mean vector
    yu = np.mean(sigmapts, axis=1, keepdims=True)

    # Compute the deviation matrix
    y = sigmapts - yu

    # Compute the covariance matrix
    cov = (y @ y.T) / n_pts

    return yu, cov



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
from typing import Tuple

import numpy as np

from ... import constants as const


# Set up logging
logger = logging.getLogger(__name__)


def pathm(llat: float, llon: float, range_: float, az: float) -> Tuple[float, float]:
    """Determines the end position (latitude and longitude) for a given range and
    azimuth from a given starting point.

    References:
        Vallado: 2022, p. 872, Eq. 11-6 and 11-7

    Args:
        llat (float): Start geocentric latitude in radians (-pi/2 to pi/2)
        llon (float): Start longitude in radians (0 to 2pi).
        range_(float): Range between points in Earth radii
        az (float): Azimuth in radians (0 to 2pi)

    Returns:
        tuple: (tlat, tlon)
            tlat (float): End geocentric latitude in radians (-pi/2 to pi/2)
            tlon (float): End longitude in radians (0 to 2pi)
    """
    # Normalize inputs
    az = az % const.TWOPI
    llon = (llon + const.TWOPI) % const.TWOPI
    range_ = range_ % const.TWOPI

    # Find geocentric latitude
    tlat = np.arcsin(
        np.sin(llat) * np.cos(range_) + np.cos(llat) * np.sin(range_) * np.cos(az)
    )

    # Find delta n, the angle between the points
    deltan = 0
    if abs(np.cos(tlat)) > const.SMALL and abs(np.cos(llat)) > const.SMALL:
        sindn = np.sin(az) * np.sin(range_) / np.cos(tlat)
        cosdn = (np.cos(range_) - np.sin(tlat) * np.sin(llat)) / (
            np.cos(tlat) * np.cos(llat)
        )
        deltan = np.arctan2(sindn, cosdn)
    else:
        # Case where launch is within a small distance of a pole
        if abs(np.cos(llat)) <= const.SMALL:
            deltan = az + np.pi if np.pi < range_ < const.TWOPI else az

        # Case where end point is within a small distance of a pole
        elif abs(np.cos(tlat)) <= const.SMALL:
            deltan = 0

    # Compute end longitude
    tlon = (llon + deltan) % const.TWOPI

    return tlat, tlon


def rngaz(
    llat: float, llon: float, tlat: float, tlon: float, tof: float = 0.0
) -> Tuple[float, float]:
    """Calculates the range and azimuth between two specified ground points
    on a spherical Earth.

    References:
        Vallado: 2022, p. 872, Eq. 11-3 to 11-5

    Args:
        llat (float): Start geocentric latitude in radians (-pi/2 to pi/2)
        llon (float): Start longitude in radians (0 to 2pi)
        tlat (float): End geocentric latitude in radians (-pi/2 to pi/2)
        tlon (float): End longitude in radians (0 to 2pi)
        tof (float): Time of flight if applicable, in seconds (default is 0)

    Returns:
        tuple: (range_, az)
            range_ (float): Range between points in km
            az (float): Azimuth in radians (0 to 2pi)
    """
    # Calculate the spherical range
    range_ = np.arccos(
        np.sin(llat) * np.sin(tlat)
        + np.cos(llat) * np.cos(tlat) * np.cos(tlon - llon + const.EARTHROT * tof)
    )

    # Check for special cases where range is 0 or half the Earth
    if abs(np.sin(range_) * np.cos(llat)) < const.SMALL:
        az = np.pi if abs(range_ - np.pi) < const.SMALL else 0
    else:
        az = np.arccos(
            (np.sin(tlat) - np.cos(range_) * np.sin(llat))
            / (np.sin(range_) * np.cos(llat))
        )

    # Adjust azimuth if it is greater than pi
    if np.sin(tlon - llon + const.EARTHROT * tof) < 0:
        az = const.TWOPI - az

    return range_ * const.RE, az


def satfov(
    az: float,
    slatgd: float,
    slon: float,
    salt: float,
    tfov: float,
    etactr: float,
    tol_fov: float = 1e-5,
) -> Tuple[float, float, float, float]:
    """Finds parameters relating to a satellite's field of view (FOV).

    References:
        Vallado: 2022, p. 874-876, Eq. 11-8 to 11-13

    Args:
        az (float): Azimuth in radians (0.0 to 2pi)
        slatgd (float): Geodetic latitude of the satellite in radians (-pi/2 to pi/2)
        slon (float): Longitude of the satellite in radians
        salt (float): Altitude of the satellite in km
        tfov (float): Total field of view in radians
        etactr (float): Center where the sensor is pointed in radians
        tol_fov (float): Tolerance for the center of the FOV (default is 1e-5)

    Returns:
        tuple: (rhomin, rhomax, lat, lon)
            rhomin (float): Minimum slant range in km
            rhomax (float): Maximum slant range in km
            lat (float): Latitude of the center of the FOV in radians
            lon (float): Longitude of the center of the FOV in radians
    """
    # Satellite parameters and limiting cases
    r = const.RE + salt

    # Maximum field of view
    fovmin = etactr + tfov * 0.5
    gamma = np.pi - np.arcsin(r * np.sin(fovmin) / const.RE)  # use larger angle
    rhomax = const.RE * np.cos(gamma) + r * np.cos(fovmin)

    # Slant range
    lambda_ = np.arcsin(rhomax * np.sin(fovmin) / const.RE)
    rhomin = lambda_ * const.RE

    # Location of center of FOV
    if abs(etactr) > tol_fov:
        lat, lon = pathm(slatgd, slon, lambda_, az)
    else:
        lat, lon = slatgd, slon

    return rhomin, rhomax, lat, lon



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 1 March 2001
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
from enum import Enum
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ... import constants as const
from ...mathtime.vector import angle
from ..time.sidereal import lstime
from ..twobody.frame_conversions import perifocal_transform
from . import utils as utils


# Set up logging
logger = logging.getLogger(__name__)


class SunEventType(Enum):
    SUNRISESET = "s"
    CIVIL_TWILIGHT = "c"
    NAUTICAL_TWILIGHT = "n"
    ASTRONOMICAL_TWILIGHT = "a"


def position(jd: float) -> Tuple[np.ndarray, float, float]:
    """Calculates the geocentric equatorial position vector of the Sun.

    This is the low precision formula and is valid for years from 1950 to 2050. The
    accuaracy of apparent coordinates is about 0.01 degrees.  notice many of the
    calculations are performed in degrees, and are not changed until later. This is due
    to the fact that the almanac uses degrees exclusively in their formulations.

    Sergey K (2022) has noted that improved results are found assuming the oputput is in
    a precessing frame (TEME) and converting to ICRF.

    References:
        Vallado: 2022, p. 285, Algorithm 29

    Args:
        jd (float): Julian date (days from 4713 BC)

    Returns:
        tuple: (rsun, rtasc, decl)
            rsun (np.ndarray): Inertial sun position vector in km
            rtasc (float): Right ascension of the sun in radians
            decl (float): Declination of the sun in radians
    """
    # Julian centuries from J2000
    tut1 = (jd - const.J2000) / const.CENT2DAY

    # Mean anomaly and ecliptic longitude of the sun in radians
    _, meananomaly, eclplong = utils.sun_ecliptic_parameters(tut1)

    # Obliquity of the ecliptic in radians
    obliquity = utils.obliquity_ecliptic(tut1)

    # Magnitude of the Sun vector in AU
    magr = (
        1.000140612
        - 0.016708617 * np.cos(meananomaly)
        - 0.000139589 * np.cos(2 * meananomaly)
    )

    # Sun position vector in geocentric equatorial coordinates
    rsun = np.array(
        [
            magr * np.cos(eclplong),
            magr * np.cos(obliquity) * np.sin(eclplong),
            magr * np.sin(obliquity) * np.sin(eclplong),
        ]
    )

    # Right ascension in radians
    rtasc = np.arctan(np.cos(obliquity) * np.tan(eclplong))

    # Ensure right ascension is in the same quadrant as ecliptic longitude
    if eclplong < 0:
        eclplong += const.TWOPI
    if abs(eclplong - rtasc) > np.pi / 2:
        rtasc += 0.5 * np.pi * round((eclplong - rtasc) / (0.5 * np.pi))

    # Declination (radians)
    decl = np.arcsin(np.sin(obliquity) * np.sin(eclplong))

    return rsun * const.AU2KM, rtasc, decl


def rise_set(
    jd: float,
    latgd: float,
    lon: float,
    event_type: SunEventType = SunEventType.SUNRISESET,
) -> Tuple[float, float]:
    """Finds the universal time for sunrise and sunset given the day and site location.

    References:
        Vallado: 2022, p. 289-290, Algorithm 30

    Args:
        jd (float): Julian date (days from 4713 BC)
        latgd (float): Geodetic latitude of the site in radians (-65 deg to 65 deg)
        lon (float): Longitude of the site in radians (-2pi to 2pi) (west is negative)
        event_type (SunEventType): Type of event to calculate
                                   (default is SunEventType.SUNRISESET)

    Returns:
        tuple: (sunrise, sunset)
            sunrise (float): Universal time of sunrise in hours
            sunset (float): Universal time of sunset in hours
    """
    # Normalize longitude to -π to π
    lon = (lon + np.pi) % const.TWOPI - np.pi

    # Select the sun angle based on the kind of event
    sunangle_map = {
        SunEventType.SUNRISESET: np.radians(90 + 50 / 60),
        SunEventType.CIVIL_TWILIGHT: np.radians(96),
        SunEventType.NAUTICAL_TWILIGHT: np.radians(102),
        SunEventType.ASTRONOMICAL_TWILIGHT: np.radians(108),
    }
    sunangle = sunangle_map.get(event_type, None)
    if sunangle is None:
        raise ValueError(f"Invalid event type: {event_type}")

    # Initialize results dictionary
    results = {"sunrise": np.nan, "sunset": np.nan}

    # Loop for sunrise and sunset
    initial_guess_times = {"sunrise": 6, "sunset": 18}
    for event, jd_offset in initial_guess_times.items():
        # Initialize Julian date for the day
        jdtemp = (
            jd
            + (np.degrees(-lon) / const.DEG2HR / const.DAY2HR)
            + jd_offset / const.DAY2HR
        )

        # Julian centuries from J2000.0
        tut1 = (jdtemp - const.J2000) / const.CENT2DAY

        # Ecliptic longitude of the Sun in radians
        *_, lonecliptic = utils.sun_ecliptic_parameters(tut1)

        # Obliquity of the ecliptic in radians
        obliquity = utils.obliquity_ecliptic(tut1)

        # Right ascension and declination in radians
        ra = np.arctan(np.cos(obliquity) * np.tan(lonecliptic))
        decl = np.arcsin(np.sin(obliquity) * np.sin(lonecliptic))

        # Local hour angle
        lha = (np.cos(sunangle) - np.sin(decl) * np.sin(latgd)) / (
            np.cos(decl) * np.cos(latgd)
        )
        if abs(lha) > 1:
            logger.error("Local hour angle out of range; sunrise/sunset not visible.")
            return results["sunrise"], results["sunset"]

        lha = np.arccos(lha)
        if event == "sunrise":
            lha = const.TWOPI - lha

        # GST and UT
        gst = (
            1.75336855923327
            + 628.331970688841 * tut1
            + 6.77071394490334e-06 * tut1**2
            - 4.50876723431868e-10 * tut1**3
        ) % const.TWOPI
        uttemp = (lha + ra - gst) % const.TWOPI
        uttemp = np.degrees(uttemp) / const.DEG2HR
        uttemp = uttemp % const.DAY2HR

        # Assign to sunrise or sunset
        results[event] = uttemp

    return results["sunrise"], results["sunset"]


def in_light(
    r: ArrayLike,
    jd: float,
    earth_model: utils.EarthModel = utils.EarthModel.ELLIPSOIDAL,
) -> bool:
    """Determines if a spacecraft is in sunlight at a given time.

    References:
        Vallado: 2022, p. 312-315, Algorithm 35

    Args:
        r (array_like): Position vector of the spacecraft in km
        jd (float): Julian date (days from 4713 BC)
        earth_model (EarthModel, optional): Earth model to use (default is ELLIPSOIDAL)

    Returns:
        bool: True if the spacecraft is in sunlight, False otherwise
    """
    # Calculate the Sun's position vector
    rsun, *_ = position(jd)

    # Determine if the spacecraft is in sunlight
    return utils.in_sight(rsun, r, earth_model)


def illumination(jd: float, lat: float, lon: float) -> float:
    """Calculates the illumination due to the sun at a given location and time.

    References:
        Vallado: 2022, p. 316-317, Eq. 5-10, Table 5-1

    Args:
        jd (float): Julian date (days from 4713 BC)
        lat (float): Latitude of the location in radians
        lon (float): Longitude of the location in radians (-2pi to 2pi) (West negative)

    Returns:
        float: Luminous emmittance, lux (lumen/m²)
    """
    # Sun right ascension and declination
    _, srtasc, sdecl = position(jd)

    # Local sidereal time
    lst, _ = lstime(lon, jd)

    # Local hour angle
    lha = lst - srtasc

    # Sun elevation
    sunel = np.arcsin(
        np.sin(sdecl) * np.sin(lat) + np.cos(sdecl) * np.cos(lat) * np.cos(lha)
    )

    # Convert sun elevation to degrees
    sunel_deg = np.degrees(sunel)

    # Compute illumination using ground illumination indices
    sunillum = 0
    if sunel_deg > -18.01:
        x = sunel_deg / 90

        # Determine coefficients based on sun elevation
        # See Table 5-1 in Vallado 2022 (p. 317)
        if sunel_deg >= 20:
            l0, l1, l2, l3 = 3.74, 3.97, -4.07, 1.47
        elif 5 <= sunel_deg < 20:
            l0, l1, l2, l3 = 3.05, 13.28, -45.98, 64.33
        elif -0.8 <= sunel_deg < 5:
            l0, l1, l2, l3 = 2.88, 22.26, -207.64, 1034.30
        elif -5 <= sunel_deg < -0.8:
            l0, l1, l2, l3 = 2.88, 21.81, -258.11, -858.36
        elif -12 <= sunel_deg < -5:
            l0, l1, l2, l3 = 2.70, 12.17, -431.69, -1899.83
        elif -18 <= sunel_deg < -12:
            l0, l1, l2, l3 = 13.84, 262.72, 1447.42, 2797.93
        else:
            l0, l1, l2, l3 = 0, 0, 0, 0

        # Compute illumination
        l1 = l0 + l1 * x + l2 * x**2 + l3 * x**3
        sunillum = 10**l1

        # Clamp sunillum to valid range
        if sunillum < 0 or sunillum >= 1e4:
            sunillum = 0

    return sunillum


def in_shadow_simple(r_sat: ArrayLike, r_sun: ArrayLike) -> bool:
    """Check if the satellite is in Earth's shadow.

    References:
        Curtis, H.D.: Orbit Mechanics for Engineering Students, 2014, Algorithm 12.3

    Args:
        r_sat (array_like): Satellite position vector in km
        r_sun (array_like): Sun position vector in km

    Returns:
        bool: Whether satellite is in attracting body's shadow
    """
    # Calculate angles
    sun_sat_angle = angle(r_sun, r_sat)
    angle1 = np.arccos(const.RE / np.linalg.norm(r_sat))
    angle2 = np.arccos(const.RE / np.linalg.norm(r_sun))

    # Check line of sight (no LOS = eclipse)
    if (angle1 + angle2) <= sun_sat_angle:
        return True

    return False


def in_shadow(r_eci: ArrayLike, r_sun: ArrayLike):
    """Check if in Earth's shadow (umbra and penumbra).

    References:
        Vallado: 2022, p. 305-308, Algorithm 34

    Args:
        r_eci (array_like): ECI position vector in km or AU
        r_sun (array_like): Sun position vector in km

    Returns:
        dict: Dictionary containing the computed angles, horizon, vertical components,
              penumbra and umbra status, and distance parameters.
    """
    # Umbra/penumbra angles
    angumb = np.arctan((const.SUNRADIUS - const.RE) / const.AU2KM)
    angpen = np.arctan((const.SUNRADIUS + const.RE) / const.AU2KM)

    # Check if in umbra/penumbra
    in_umbra, in_penumbra = False, False

    if np.dot(r_eci, r_sun) < 0:
        # Get satellite's vertical and horizontal distances
        sun_sat_angle = angle(-np.array(r_sun), np.array(r_eci))
        sathoriz = np.linalg.norm(r_eci) * np.cos(sun_sat_angle)
        satvert = np.linalg.norm(r_eci) * np.sin(sun_sat_angle)

        # Calculte penumbra vertical distance
        x = const.RE / np.sin(angpen)
        penvert = np.tan(angpen) * (x + sathoriz)

        # Check if in penumbra
        if satvert <= penvert:
            in_penumbra = True
            y = const.RE / np.sin(angumb)

            # Calculate umbra vertical distance
            umbvert = np.tan(angumb) * (y - sathoriz)

            # Check if in umbra
            if satvert <= umbvert:
                in_umbra = True

    return in_umbra, in_penumbra


def cylindrical_shadow_roots(
    a: float, e: float, beta_1: float, beta_2: float
) -> np.ndarray:
    """Calculate roots of cylindrical shadow quartic equation.

    References:
        Vallado: 2022, p. 310, Equation 5-6

    Args:
        a (float): Semimajor axis in km
        e (float): Eccentricity
        beta_1 (float): First temporary parameter
        beta_2 (float): Second temporary parameter

    Returns:
        np.ndarray: Roots of cylindrical shadow model
    """
    alpha = const.RE / (a * (1 - e**2))

    # Shadow coefficients
    a0 = (
        alpha**4 * e**4
        - 2 * alpha**2 * (beta_2**2 - beta_1**2) * e**2
        + (beta_1**2 + beta_2**2) ** 2
    )
    a1 = 4 * alpha**4 * e**3 - 4 * alpha**2 * (beta_2**2 - beta_1**2) * e
    a2 = (
        6 * alpha**4 * e**2
        - 2 * alpha**2 * (beta_2**2 - beta_1**2)
        - 2 * alpha**2 * (1 - beta_2**2) * e**2
        + 2 * (beta_2**2 - beta_1**2) * (1 - beta_2**2)
        - 4 * beta_1**2 * beta_2**2
    )
    a3 = 4 * alpha**4 * e - 4 * alpha**2 * (1 - beta_2**2) * e
    a4 = alpha**4 - 2 * alpha**2 * (1 - beta_2**2) + (1 - beta_2**2) ** 2

    return np.roots([a0, a1, a2, a3, a4])


def eclipse_entry_exit(
    r_sun: ArrayLike,
    a: float,
    e: float,
    i: float,
    raan: float,
    w: float,
    adjust: bool = False,
    tol: float = 1e-6,
) -> Tuple[float, float]:
    """Calculate eclipse entry and exit angles using the cylindrical shadow model.

    References:
        Vallado: 2022, p. 309-311
        Falck, R., and Dankanich, J.: Optimization of Low-Thrust Spiral Trajectories
            by Collocation, AIAA 2012-4423

    Args:
        r_sun (array_like): Sun position vector in km
        a (float): Semimajor axis in km
        e (float): Eccentricity
        i (float): Inclination in radians
        raan (float): Right ascension of ascending node in radians
        w (float): Argument of periapsis in radians
        adjust (bool, optional): Whether to adjust angles so that entry > exit
                                 (default is False)
        tol (float, optional): Tolerance for root finding (default is 1e-6)

    Returns:
        tuple: (theta_en, theta_ex)
            theta_en (float): True anomaly of shadow entry in radians
            theta_ex (float): True anomaly of shadow exit in radians
    """
    # Initialize entry/exit angles
    theta_en, theta_ex = np.pi, -np.pi

    # PQW transformation vectors (just P & Q used here)
    p_, q_, _ = perifocal_transform(i, raan, w)

    # Semiparameter and alpha
    p = a * (1 - e**2)
    alpha = p**2 / const.RE**2

    # Temp parameters
    beta_1 = np.dot(r_sun, p_) / np.linalg.norm(r_sun)
    beta_2 = np.dot(r_sun, q_) / np.linalg.norm(r_sun)

    # Solve quartic equation roots
    roots = cylindrical_shadow_roots(a, e, beta_1, beta_2)

    # Check for complex roots and return zeros if encountered
    if any(isinstance(r, complex) for r in roots):
        logger.debug(
            f"Roots are complex for OE: \n"
            f"a = {a} km, e = {e}, i = {np.degrees(i)} deg, "
            f"raan = {np.degrees(raan)} deg, w = {np.degrees(w)} deg"
        )
        return theta_en, theta_ex

    # Check for validity of all roots to remove any false positives
    # There are 8 potential solutions from the roots (Falck and Dankanich)
    thetas_good = []
    for theta in np.arccos(roots):
        # Iterate over possible sine values to get other 4 roots
        for y in [np.sin(theta), -np.sin(theta)]:
            # Pass the roots through the shadow function
            # A valid root will return ~0 (there should be 4 of these total)
            sol = np.arctan2(y, np.cos(theta))
            s = (
                (1 + e * np.cos(sol)) ** 2
                + alpha * (beta_1 * np.cos(sol) + beta_2 * np.sin(sol)) ** 2
                - alpha
            )

            # Then check that the sun angular separation is >= 90 deg
            # i.e. the cosine of the separation angle is <= ~0
            check = beta_1 * np.cos(sol) + beta_2 * np.sin(sol)

            # Check both conditions
            if abs(s) < tol and check <= tol:
                thetas_good.append(sol)

    # Check if true anomaly is shadow entry or exit
    for theta in thetas_good:
        # Calculate shadow derivative
        ds_dtheta = 2 * (p**2 / const.RE**2) * (
            beta_1 * np.cos(theta) + beta_2 * np.sin(theta)
        ) * (-beta_1 * np.sin(theta) + beta_2 * np.cos(theta)) - 2 * e * np.sin(
            theta
        ) * (
            1 + e * np.cos(theta)
        )
        if ds_dtheta > 0:
            theta_en = theta
        elif ds_dtheta < 0:
            theta_ex = theta
        else:  # pragma: no cover
            logger.warning(
                f"Unable to determine shadow entry/exit for true anomaly of "
                f"{np.degrees(theta): .2f} deg"
            )

    # Adjust angles if exit angle > entry angle (for integration)
    if adjust and theta_ex > theta_en:
        theta_en += const.TWOPI

    return theta_en, theta_ex
